const char *version_string="2019-05-26";
// ToDo: use _OPENMP macro autoset by -fopenmp (grep -e omp -e openmp *.c)
// // #define CFG_OMP  /* obsolete! use: gcc -fopenmp ... */
// #define CFG_MPI        /* define CFG_MPI + mpicc ... */ 
// #define CFG_CPUSET     /* options --pin_ncpu=... */
#if defined(_OPENMP) && !defined(CFG_OMP)  /* 2016-01-14 */
# define CFG_OMP
#endif
/*
 measure speed of memory (best and worst case, and CPU + FLOP:Byte-Ratio=ToDo)
 also FDIR
  - Ausfaelle erkennen (fault-detection)          
  - Ausfaelle einkreisen und bestimmen (fault-isolation)
  - geeignete Korrekturen (recovery)    
 
 INST_SEQ: # single task (no parallelism)
          gcc -Wall -O2 -ftree-vectorize -funroll-loops -o memspeed memspeed.c
          # memory footprint: 
 INST_MPI: # OpenMPI-usage (parallel on distributed memory)
          OMPI_CC=gcc492 mpicc -DCFG_MPI -Wall -O2 \
           -ftree-vectorize -funroll-loops \
           -o memspeed memspeed.c
          # optional: -DCFG_CPUSET=0x0ff -fopenmp
          #   cpuset is a hex bit pattern for used cpu cores (p.e. 0x05,0x0f)
          #     useful to bind tasks to cores using sched_setaffinity()
          #   use cc or gcc if mpicc is not availabe
          mpirun -bind-to-core -np <tasks>\
            ./memspeed --asize=<MBsize> --mintime=<sec>
          # needs: memory = tasks*2^log2size Bytes
          #        time   = (1..2)*2*mintime*(log2size-8) seconds
          mpirun -np 8 --bynode ./memspeed --log2size=29 #    4GB=8*512MB
          mpirun -np 2 --bynode ./memspeed --log2size=27 #  256MB=2*128MB
          mpirun -np 2 --bynode ./memspeed --asize=384   #  768MB=2*384MB
          # runtime is about 2*(minsize...2minsize)*(log2size-7)
          ... --pin_ncpu=6 ...
 INST_OMP:# OpenMP-usage (parallel on shared memory)
          yum install libgomp
          gcc -Wall -O2 -fopenmp -o memspeed memspeed.c
          export OMP_WAIT_POLICY="PASSIVE" # OpenMP-3.0 waits saving power 
          export OMP_NUM_THREADS=8         # OpenMP     8 threads
          OMP_PROC_BIND=TRUE # thread bounding, thread affinity
          OMP_WAIT_POLICY="PASSIVE" OMP_NUM_THREADS=2 ./memspeed
          # memory footprint: 
 SLURM:   # job-system
          # task/affinity or task/cgroup must be enabled, scontrol show config
          # SLURM_CPU_BIND=verbose # -c = --cpus-per-task (OpenMP) dflt=1
          srun -N1 -c4 --cpu_bind=verbose,mask_cpu:0xf,0xf0,0xf00,0xf000 ./memspeed --asize=384
          srun -N1 --ntasks-per-node=4 --cpu_bind=verbose,mask_cpu:0xf,0xf0,0xf00,0xf000
          # -u = --unbuffered (output)
          #  SLURM_CPU_BIND_VERBOSE SLURM_CPU_BIND_TYPE SLURM_CPU_BIND_LIST is set
          # srun works with mpich, for openmpi mpirun musst be called:
          srun -N1-2 -c8 bash -c "[ \$SLURM_PROCID == 0 ]&&mpirun ./memspeed||true"
          srun -p sw01_short -N1-2 -c8 bash -c "[ \$SLURM_PROCID = 0 ]&&OMP_NUM_THREADS=4 mpirun ./memspeed --test=1 || true"
 MONITOR: ps -Lf     # UID PID PPID LWP(!) C NLWP STIME TTY TIME CMD
          ps -T      # PID SPID TIME TTY CMD
          top -H
          cat /sys/devices/system/edac/mc/mc[0-3]/ce_count | xargs # ECC-Errors
          cat /sys/devices/system/edac/mc/poll_msec # 1000 ms # x86+amd
 TEST-Suites:
          kautz2009-2012 lot of ECC-Errors and types (CE,UE,panics)
          comp2-ECC-CEs+Crashs

ToDo: better read random index from array?
 - use much faster random_xorshift()
 - random_update vy[prefetchable_rnd_iy]+=nzHxy[rnd_iy]*vx[seq_ix]
   from rnd_nz_matrix p.e. NZX=40; vx,vy,sym_sparse_matrix[iy,ry] SpMV
  CSR = compressed sparse row, 
   large.N>=1e5 nnz/row=10..100 100-300MFLOP/s
  Benchmarking_Sparse_Matrix_Vector.pdf
    diag1: x=N y=nnz/row color=MFLOP/s
    diag2: x=N y=MFLOP/s color=nnz/row(fix to log(N)?)
 - use vector-code (SSE2, AVX2) for speedup
 - reduce code size and memory footprint (90M-180M openmpi), mpi-less?
 - check FLOPs for mem+power stress using oprofile
 - modules speed random, random_prefetch, seq, r[w], + stress + mark badmem
   MEMstress +FPU+GPU+DSK20W+DVD+USB2W+ETH+IB(=PSU)-stress? vs. Verbrauch?
   test: stresstypes vs. power on comp2
   see badRAM-linux-patch + memtest86-tests(random+mod20)
    output speed only if changed, not after 10 changes (less output) +timestamps
 - add OpenMP threads (alternative: fork/clone+pipe or pthreads)
ChangeLog:
 2014-06-26: more help, NA controls prefetching of random access now
 2013-07-02: fix problem with i(neg.long) % a2 becoming 0xfff... -gt a2
 2013-06-18: improve on 972*4GB cluster which has bitflips (kautz)
 2013-06-14: option --asize=8192 (units of log2size.default=20 fror MB)
             ini mem, to avoid page misses on 1st loop, lstream.p4
 2013-06-06: --test=2 prefetch random + --cpu_...=...
 2013-05-02: use wtime [sss.mmmuuu], not itime [sss.000000] if not MPI
             add CFG_MPI + -fopenmp, add test8(r+w)=stresstest_mem+psu
 2008-08-22: add mpi-functionality to fill all cores of a node,
             add a random shaked pointer chain (less instructions)
  SRC-RND: do { ll=aa[ll]; } while(--k);
  SRC-STR: do { x^=*ap; ap++; } while(--k);
  ; gcc-4.3.3-x86-64 .L41 = 3 instr/loop, .L59 = 4 instr/loop
  .L52: subl    $1, %eax             ; gcc-4.1.1-i386   3 instr/loop
        movl    (%edx,%ebx,4), %ebx
        jne     .L52
  .L50: subq    $1, %rax             ; gcc-4.1.0-x86-64 3 instr/loop
        movq    (%rdx,%rbp,8), %rbp
        jne     .L50
  .L55: ... shladd r14 = r37, 3, r15  ; gcc-4.1.1-ia64  3 instr/loop
        ... ld8    r37 = [r14]
        ... br.cloop.sptk.few .L55
  L$66: s8addq  $10, $0, $10          ; cxx-6.5-alpha 19 instr/8loop
        lda     $2, -8($2)            ;  + 5 inst/loop 
        ldq     $10, ($10)
        cmple   $2, 7, $3
        s8addq  $10, $0, $10
        ldq     $10, ($10)  
        ... above 2 lines repeated 7 times (unrolled loop)
        beq     $3, L$66
                                        
 2007-04-27:  4 x86-instructions per loop (gcc-4.1.1 -O2)
              5 opteron-Instr.   per loop (gcc-4.1.0 -O2)
              4 sparcv9-Instr.   per loop (gcc-4.1.1 -O2)
             32 alpha-Instr.     per loop (cxx-6.5 -fast)
  .L15: xorl    aa(,%edx,4), %edi    ; gcc-x86-p4
        addl    %ebx, %edx
        subl    $1, %eax
        jne     .L15
  .L16: ldx     [%g3], %g1           ; gcc-sparcv9
        add     %g2, -1, %g2
        xor     %l0, %g1, %l0
        brnz,pt %g2, .L16   
  .L900000128:                       ; CC-5.3-sparcv9
      add     %o1,%i3,%o1
      xor     %l0,%o2,%l0
      add     %i2,%i4,%i2
      addcc   %o4,-1,%o4
      bne,a,pt        %xcc,.L900000128

 # P4-2.6GHz/2_maxspeed=(1.3GHz/(loop=5Instr)=260MHz)*(long=4)=1040MB/s
 # see also memspeed.gpl
 # compare with numademo 8000000000 # 8GB per run? (drop_caches! 5 vs. 8GB/s)
 #    2012comp2 memset=5-8GB/s 
 # System               Array Rand MaxMB/s Err  # new 4x->4summed
 Nehalem-2GHz-4MB       256MB   70   6338       114ns 64bit-gcc-4.3.3 72GB -O2
 Nehalem-2GHz-4MB       256MB   70   6340       114ns 64bit-gcc-4.3.3 72GB -O1
 Nehalem-2GHz-4MB       256MB   68   1850       118ns 64bit-gcc-4.3.3 72GB -O0 more x86-instr.
 Nehalem-2GHz-4MB    2x 256MB   66   4707       122ns 64bit-gcc-4.3.3 72GB -O0 more x86-instr.
 PM- 600MHz-1MB         512kB  284    393    -   14ns 32bit-gcc-4.1.2 2GB hdparm=385MB/s 0808
 PM- 600MHz-1MB         256MB   23    382    -  174ns 32bit-gcc-4.1.2 2GB hdparm=385MB/s 0808
 PM-1400MHz-1MB          64MB   31    870    -  127ns 32bit-gcc-4.1.2 2GB
 P4-2.6GHz/2-512kB      256MB   20   1090    -  197ns 32bit-gcc-4.1.1 hdparm=1181MB/s FSB=4*100MHz Mem=266MHz*8=2128MB/s CPU=13*100MHz (default=13*200MHz) L2=512k=2^19 0808
 P4-2.6GHz/2-512kB HT 2x256MB   26   1533    -  307ns 32bit-gcc-4.1.1 hdparm=1181MB/s FSB=4*100MHz Mem=266MHz*8=2128MB/s CPU=13*100MHz (default=13*200MHz) L2=512k=2^19 0808
 SunFire-880             32MB   23    220    6  346ns 64bit-gcc-4.1.1 USparcIII 750MHz
 SunFire-880             32MB   23    232    7  333ns 64bit-CC-5.3-fast-xarch=v9
 SunFire-V490            32MB   92    798   19   84ns 64bit-gcc-4.1.1 USparcIV+ 10*150MHz L2=32MB 8-way
 DOpteron275-2.2GHz      32MB   89   2130   34   88ns 64bit-gcc-4.1.1 L2=1MB  hdparm-T=1945MB/s 2.6.18 2DualCore
 DOpteron275-2.2GHz  2x  32MB  177   4129   16   88ns 64bit-gcc-4.1.1 L2=1MB  hdparm-T=1945MB/s 2.6.18 2DualCore 4*44MB/s+2*85MB/s
 DOpteron275-2.2GHz  4x  32MB  340   7156   92   91ns 64bit-gcc-4.1.1 L2=1MB  hdparm-T=1945MB/s 2.6.18 2DualCore
 DOpteron275-2.2GHz  6x  32MB  346   7200   98  177ns 64bit-gcc-4.1.1 L2=1MB  hdparm-T=1945MB/s 2.6.18 2DualCore 4*44MB/s+2*85MB/s
 DOpteron275-2.2GHz  8x  32MB  332   7248   92  189ns 64bit-gcc-4.1.1 L2=1MB  hdparm-T=1945MB/s 2.6.18 2DualCore
 DOpteron885-2.6GHz      32MB   35    838   21  223ns 64bit-gcc-4.1.0 8user   hdparm-T=1660MB/s BW=6.4GB/s(1CPU)
 4HOpteron8431-2.4GHz 8x 32MB 8*79 8*1820    -  101ns 64bit-gcc-4.1.2 3user   hdparm-T=5800MB/s (summed)
 Xeon-2333Mz-4MB         32MB   69   3295       116ns 64bit-gcc-4.1.1-long thor model=15 stepping=4  i5000P (maxBW=21GB/s) 2Quad
 Xeon-2333Mz-4MB     4x  32MB 4*62 4*1466       131ns 64bit-gcc-4.1.1-long thor model=15 stepping=4  i5000P (maxBW=21GB/s) 2Quad
 Xeon-2333Mz-4MB     8x  32MB 8*55  8*734       144ns 64bit-gcc-4.1.1-long thor model=15 stepping=4  i5000P (maxBW=21GB/s) 2Quad 
 Xeon-2333Mz-4MB     1x 512MB   60   3269       133ns 64bit-gcc-4.1.1-long thor model=15 stepping=4  i5000P (maxBW=21GB/s) 2Quad
 Xeon-2333Mz-4MB     8x 512MB 8*39  8*735       204ns 64bit-gcc-4.1.1-long thor model=15 stepping=4  i5000P (maxBW=21GB/s) 2Quad  Aug08
 alpha-gs160-731MHz      16MB   20    295   49  397ns 64bit-gcc-4.1.1-long
 alpha-gs160-731MHz      16MB   20    295   49  397ns 64bit-cxx-6.5-long
 alpha-gs1280            32MB   31    453   50      - 64bit-cxx-6.5-long   (32cpus_per_rad) 0user
 alpha-gs1280            32MB   46    644   22  170ns 64bit-cxx-6.5-long   (8cpus_per_rad) 8user
 alpha-gs1280            32MB   46    339  170  170ns 64bit-gcc-4.1.1-long (8cpus_per_rad) 8user
 alpha-gs1280            32MB   77   1109    9  101ns 64bit-cxx-6.5-long   (1cpus_per_rad) 0user
 alpha-gs1280         32x32MB 2432  14400 1600      - 64bit-cxx-6.5-long   (32cpus_per_rad) 0user
 alpha-gs1280         32x32MB 2432  35520  640  102ns 64bit-cxx-6.5-long   (1cpus_per_rad) 0user
 alpha-es45-1250MHz      32MB   38    462   56  204ns 64bit-cxx-6.5-long 4user
 altix330-IA64-1.5GHz    32MB   22    309   12  349ns 64bit-gcc-4.1.1 Apr07
 altix330-IA64-1.5GHz 8x 32MB  168   2304   80  970ns 64bit-gcc-4.1.1 Apr07
 altix330-IA64-1.5GHz 15x32MB  660   9300  300  176ns 64bit-gcc-4.1.1 Apr07
 altix330-IA64-1.5GHz    32MB   54    670   14  143ns numactl --localalloc
 altix330-IA64-1.5GHz 8x 32MB  392   4760  360  158ns numactl --localalloc
 altix330-IA64-1.5GHz 15x32MB  735  10050  300  158ns numactl --localalloc
 altix330-IA64-1.5GHz 8x512MB  418   5750       153ns min= 8*52MB/s aff+local
 altix330-IA64-1.5GHz 8x128MB  427   5748       150ns min= 8*53MB/s aff+local
 altix330-IA64-1.5GHz 2x128MB   91    709       174ns min= 2*45MB/s c0,1+local
 altix330-IA64-1.5GHz 2x128MB  107   1438       150ns min= 2*53MB/s aff0,8
 altix330-IA64-1.5GHz 2x512MB   89    710       179ns min= 2*45MB/s c0,1+local
 altix330-IA64-1.5G  16x128MB  448   6928       285ns min=16*28MB/s localalloc
 altix330-IA64-1.5G  16x128MB  700  10800       182ns min=16*44MB/s localalloc
 altix330-IA64-1.5G  16x128MB  729  11335       182ns min=16*45MB/s aff+lalloc
 altix330-IA64-1.5G  16x512MB  480   7456       264ns min=16*30MB/s localalloc
# echo 1 > /proc/sys/vm/drop_caches + numademo 1024m # ???
# 30*8quantum numademo 1024m 1node  stream=3.7GB/s     rnd=137MB/s
# 22*16comp2  numademo 1024m 4nodes stream=2.8-6.3GB/s rnd=93-125MB/s 4*16GB
# check with linkstat!
# ----------- old data (step only) --------------------
 System                 Array Min MaxMB/s Err ln2MinStep # ---- old(!) ----
 PM- 600MHz-1MB 512MB    16MB   52    364   16    12   
 PM-1400MHz-1MB 512MB    16MB   50    364    8    12   
 Xeon-3GHz-1MB  2GB     256MB   39   1800   48    16   32bit-gcc-2.96  connect model=3 stepping=4  hdparm=800MB/s  busclock=200MHz
 Xeon-3GHz-2MB  2GB     256MB   44   2700   57    20   32bit-gcc-4.0.1 bb    model=4 stepping=3  hdparm=1472MB/s
 Xeon-3GHz-2MB 12GB     256MB   39   1422   31    20   64bit-gcc-4.1.1 morus model=4 stepping=10 memtest86+1.65=1257MB/s hdparm=1286MB/s
 Xeon-3GHz-2MB 12GB     512MB   38   1450   31    20   32bit-gcc-4.1.1-long morus model=4 stepping=10 memtest86+1.65=1257MB/s hdparm=1286MB/s
 Xeon-3GHz-2MB 12GB    8192MB   80   1170   23    19   64bit-gcc-4.1.1-long morus model=4 stepping=10 memtest86+1.65=1257MB/s hdparm=1286MB/s
 alpha-gs1280 128GB      32MB   36   1117   39    15   64bit-gcc-3.4.3-long (32cpus_per_rad) 2/82ns==25MB/s
 alpha-gs1280 128GB      32MB   41   2485   45    15   64bit-cxx-6.5-long   (32cpus_per_rad) (32x2.5GB/s=80GB/s local) remote.maxBW=6.3GB/s local.maxBW=12.3GB/s(767MHz*8channels*2Byte) 8B/(82ns..250ns)=>32..98MB/s
 alpha-gs1280 128GB      32MB   67   1117   39    16   64bit-gcc-3.4.3-long (1cpus_per_rad) 2/82ns==25MB/s
 alpha-gs1280 128GB      32MB   61   2485   47    16   64bit-cxx-6.5-long   (1cpus_per_rad) (32x2.5GB/s=80GB/s local)
 alpha-gs1280 128GB      16GB   24    572    7    23   64bit-cxx-6.5-long   (8cpus_per_rad)
 alpha-gs1280 128GB      16GB   29   1094   24    23   64bit-gcc-4.1.1-long (8cpus_per_rad)
 alpha-es45-1250MHz      16MB   87    818   20    11   cxx-int? 4/5CPU
 DOpteron275-2.2GHz-1MB  16MB   35   1500   34    13   64bit-gcc-3.4.6      hdparm-T=1900MB/s
 DOpteron885-2.6GHz-1MB 256MB   38   1120   39    15   64bit-gcc-4.0.3      hdparm-T=1660MB/s BW=6.4GB/s(1CPU)
 DOpteron885-2.6GHz-1MB 512MB   68   2300   82    15   64bit-gcc-4.0.3-long hdparm-T=1660MB/s
 DOpteron885-2.6GHz-1MB 512MB*4 46   2220   77    15   CPU3-6! 64bit-gcc-4.0.3-long hdparm-T=1660MB/s
 MinMax fuer 2 Jobs etc. # linkstat?
 cpu.model=generation stepping=version (same masque)
 ToDo: try __*_prefetch and improve loop (tune for generated asm-code)
 Theory vs. Samples (yum install edac-utils):
   NumCPUs*Memchannels/CPU*maxDatarate vs. DIMMs/Channel*Ranks/DIMM*Speed
 see ct23/2013 REP MOVSB different speeds, bugs (break loop, INT,VMs,...)
 Memory-Footprint: openmpi is bad! 128MB/node (tune ompi-compilation!)
 comp2: 2*AMD-6134*4Channels * max 3 DIMMs/channel = 2*42.7GB/s bei 1DIMM/ch
        2*8GB/Channel (DDR3-1333) = 10.664GB/s / channel = 85.312GB/s
          aBW=22.3GB/s? (unroll2?=26.0GB/s)  memtest4.20=3746MB/s(singleCPU)
          aBW=46.6GB/sR 2013-06
           http://en.wikipedia.org/wiki/Memory_footprint
           1MBmemusage: approx 128MB sm for openmpi
            --mca btl self,sm,tcp # 1*6M/126M ... 4*6M/191MB... 8*7M/193MB
              67MB tmp/openmpi-sessions-jschulen@comp2_0/60576/1/shared_mem_pool.comp2
            --mca btl self,tcp # 8*5M/91MB (9MB openmpi-libs)
            single-task-mpi: 2M/23MB 9MB-ompi-lib
            single-task-nompi: 1.5M/5MB 2MB-libc /proc/$PID/smaps
     worst random at mem/core/2 = 4GB 
        test 1GB: 1*=40ns/8B 2*  16*155MB/s=52ns/8B/16 ca. 40 cycles
        test 4GB:                16*136MB/s=59ns/8B/16
 quantum: 2*E5450 (8GB/s?)
          8 DDR2-667(PC-5300F)=5.3GB/s(10.6GB/s DualChannel)
          aBW=6.17GB/sRW   
 pse:   2*6core-iX5650-2.66GHz = 2*3 Speicherkanaele(DDR3-1333) max 2*32GB/s
        18*16GB-DDR3-1066MHz(a8.5GB/s) = 3 DIMMs/Channel max 51GB/s (ToDo)
          aBW=29.5GB/sRW 
 quanta1: 4*8 L1=32kB(0.5-4ns) L2=4*8*256kB(0.7-10ns) L3=16MB(1.05-33ns) mem(1.44-128ns)
          aBW1=5.5GB/s rand1=63MB/s(127ns,760MB)
  # Tasks Mem/Task[B] loops t[s]  aBW[B/s] BW[B/s] lat[ns]
  1 7.73e+11         7 3110.84 3.48e+09 3.48e+09   4.60 lr+w4*+16
  1 7.73e+11         7 3080.90 3.51e+09 3.51e+09   4.55 lr+w4*+16
  1 7.73e+11         7 3042.04 3.56e+09 3.56e+09   4.50 lr+w4*+16
ToDo:
 - include memtest.c
 - add pthread
 - memory-footprints (pure,mpi-???,ompi-???,stst,dyn)
  memusage(ELF64+libc6+1MB+test32) method: (ulimit -v ....;./a.out) 2013-09
  gcc4.1.2                   4.8MB       
  gcc4.1.2+static            1.9MB       
  gcc4.1.2+openmp2.5        11.2MB(1) 21.5MB(2) 31.7MB(3) 42.0MB(4) 82.9MB(8)
  gcc4.1.2+openmp2.5-static  2.0MB(1) 12.2MB(2) 22.4MB(3) 32.7MB(4) 73.7MB(8)
  openmpi2.1+gcc4.1.2        90MB(1)   n*155MB(2-8)      
  gcc4.1.2+-lpthread         6.9MB(1) 
  # RESUME2013: -fopenmp +10MB*n overhead! ompi=+154MB*n overhead
 - messe auch free-difference or max malloc (shared LIBS!)
 - einzelprogramme besser als OMPI!?
 - audiotxt verbessern arecord als stream! fuer const. mem on PC
         
*/


#define _GNU_SOURCE 1 // this defines __USE_GNU in features.h for sched.h

#ifdef _OPENMP
#include <omp.h>    /* yum install libgomp; gcc -fopenmp ... */
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h> /* 2013-05 adding strncmp() */
#include <time.h>
#include <unistd.h> /* usleep */

// TODO alignement
// struct S { short f[3]; } __attribute__ ((aligned (8)));
// typedef int more_aligned_int __attribute__ ((aligned (8)));
// struct S { short f[3]; } __attribute__ ((aligned)); // max aligned
//   sometimes linkers can to go above some (p.e. 8 Byte) alignment
//   alignment may be ignored
    

#ifndef CFG_MPI    // for the case of MPI is not available
/* last tests: 2013-05-02 */
#include <sys/time.h>
double wtime(){ struct timeval tv;
 if(gettimeofday(&tv, NULL)){perror("timeofday");exit(2);}
 return tv.tv_sec+1e-6*tv.tv_usec; }
#define MPI_Init(a,b) 1
#define MPI_SUCCESS 1
#define MPI_Comm_size(MPI_COMM_WORLD, mpi_n)
#define MPI_Comm_rank(MPI_COMM_WORLD, mpi_i)
#define MPI_Barrier(MPI_COMM_WORLD)
#define MPI_Wtime()  wtime() /* 2013-05-02 itime replaced by wtime */
#define MPI_Finalize()
#define MPI_Abort(MPI_COMM_WORLD,xx) exit(xx)
#define MPI_Bcast(pt, len, MPI_DOUBLE, id, MPI_COMM_WORLD)
/* Reduce replacement works for MIN MAX SUM,
   Reduce to rank0, Allreduce to all ranks */
#define MPI_Reduce(snd, rec, SIZE, M_INT, M_MIN, rk, MCW) *rec=*snd
#define MPI_Allreduce(snd, rec, SIZE, M_INT, M_MIN, MCW) *rec=*snd
#else  // CFG_MPI
#include <mpi.h>
#endif // CFG_MPI

#ifdef CFG_CPUSET // switch on if sched_setaffinity will benefit
#include <unistd.h>  // declare sleep()+usleep
#include <sched.h>   // declare sched_setaffinity()
#ifndef _FEATURES_H  
# warning "_FEATURES_H is not defined, try top add #include <features.h>"
#endif // _FEATURES_H
#ifndef __USE_GNU
// bcc 0.16.17 does not like #warning?
# warning "expect __USE_GNU defined, but it isn't, try -D_GNU_SOURCE=1"
#endif // __USE_GNU
#endif // CFG_CPUSET


long int *aa, alen;  // long array and size of it

void check_ring( long int a2, int vvv, int mpi_i ) {
  long int k, l;
  // check that we have still a full ring
  if (vvv && !mpi_i) printf("# chain:");
  for (l=0,k=0; k<a2; k++) {
    if (vvv && k <8 && !mpi_i) printf(" %2d", (int)l);
    if (vvv && k==8 && !mpi_i) printf(" ...");
    l=aa[l];
    if (k<a2-1 && l==0)
      printf("\n# ring stops early at %ld %.2f%%\n ...", k, k*100./a2);
  } if (vvv && !mpi_i) printf("\n");
  return;
}


__attribute__ ((hot))
int main(int argn, char *argv[]){
  double t0, t1, t2, t3, tmin=1.0;
  int i1, 
      omp_n=1,          /* OpenMP omp_num_threads() */
      mpi_n=1, mpi_i=0, /* MPI threads and thread number */
      pin_flag=0, /* thread to processor pinning */
      testmode=  5, /* 1=seq_read 2=pf_random 4=nonpf_random 8=stress ... */
      log2size= 20, // 2^20=1MB  set by argument1 (old, new: unit)
      log2long;     // 5 for 2^5=int32 or 6 for 2^6=int64 (long int *aa)
  int opt_time=10;  // walltime dflt=10s (max=2e9s=68y) for test=32 (10s)
  long int loops, asize= 1 /* new: uses log2size as unit */,
     i, j, r00, r01, r02, r10, r11, r12, a2, k,
     a_rnd[1024], /* array of random numbers to speedup randomization */
     ll=0, xy=0; // dummy vars for the innerst loop (unique names!)
  long int memhole0=0, memhole1=0; // memory hole (bad memory) 2013-11-07
  if (MPI_Init(&argn, &argv) != MPI_SUCCESS) exit(1);
  MPI_Comm_size(MPI_COMM_WORLD, &mpi_n);
  MPI_Comm_rank(MPI_COMM_WORLD, &mpi_i);
  setvbuf(stdout, NULL, _IONBF, 0); /* dont buffer output */
  if (!mpi_i)
   printf("# memspeed %s (c) 2008-2019 GPLv2 Joerg Schulenburg ovgu.de\n",
           version_string);
  /* 2013-05 add option strings --... for more flexibility */
  for (i1=1;i1<argn;i1++) { /* check all args */
    if (strncmp(argv[i1],"--asize=",8)==0) { asize=atoi(argv[i1]+8);
      if (asize<1) asize=1; continue; } // min 1 * 2^log2size 
    if (strncmp(argv[i1],"--log2size=",11)==0) { log2size=atoi(argv[i1]+11);
      if (log2size<10) log2size=10; continue; } // min 1K 
    if (strncmp(argv[i1],"--mintime=",10)==0) { 
      tmin=atof(argv[i1]+10); if (tmin<0.0) tmin=0.0; continue; }
    if (strncmp(argv[i1],"--test=",7)==0) { 
      testmode =atoi(argv[i1]+7); continue; } // testmode 32=CPU
    if (strncmp(argv[i1],"--time=",7)==0) { 
      opt_time =atoi(argv[i1]+7); continue; } // walltime
    if (strncmp(argv[i1],"--memhole=",10)==0) { int i2;
      memhole0 =atoi(argv[i1]+10);  // MB-units (1<<20)! from
      for(i2=10;argv[i1][i2]!=0
            &&  argv[i1][i2]!='-';i2++);
      if ('-'== argv[i1][i2]) memhole1= atoi(argv[i1]+i2+1); // to (exclusive)
      continue; }
    /* pin_* = processor core/thread pinning */
    if (strncmp(argv[i1],"--pin_ncpu=",11)==0) { pin_flag=1; continue; }
    if (strncmp(argv[i1],"--pin_ofs=", 10)==0) { pin_flag=1; continue; }
    if (strncmp(argv[i1],"--pin_dist=",11)==0) { pin_flag=1; continue; }
    if (strncmp(argv[i1],"--pin_size=",11)==0) { pin_flag=1; continue; }
    if (!mpi_i) {
     printf("ERROR: bad or unknown options or arguments\n"
           " use: ./memspeed --asize=16 --log2size=20 --mintime=1 --test=5\n");
     fputs("  --pin_ncpu=16 --pin_ofs=0 --pin_dist=1 --pin_size=1\n",stdout);
     fputs("  --memhole=14-16   # hole of 2MB at 14MB (bad mem)\n",stdout);
     fputs("# --test=1    = sequential read (lstream)\n",stdout);
     fputs("# --test=2    = NA-prefetchable random read\n",stdout); /* pf_ */
     fputs("# --test=4    = nonprefetchable random read\n",stdout); /* */
     fputs("# --test=8|16 = power stress\n",stdout);
     fputs("# --test=32   = power stress, max FLOPs (noMem)\n",stdout);
     fputs("# --time=10   = walltime in seconds\n",stdout);
    }
    usleep(100000l); /* give 100ms for output */ MPI_Abort(MPI_COMM_WORLD,1);
  }
  /* get the log2size of long int, log2long-3 is log2(Bytesize(long)) */
  for (log2long=0; 1<<log2long < 8*sizeof(long int); log2long++);
  alen = asize<<(log2size-log2long+3); /* log2(long=int32): log2long=5 */
  asize= asize<< log2size;

#if defined(CFG_CPUSET) && defined(__USE_GNU)
  if (pin_flag==1) { /* processor/thread pinning */
   int j1;
   // ToDo: openmp code missing
   // this processor bounding is useful for the case, that the jobsystem
   //   does not do this automatically
   // CFG_CPUSET is the bitmap for available SMP-CPUs per node
   // CFG_CPUSET=0x0005: use core 0 and 2 (Nodes with 2 Hyper-Threading CPUs)
   //   ToDo: define on runtime per taskset + sched_getaffinity ? (jobsystem?)
   // use --byslot for numbits(CFG_CPUSET)<mpi_n
   cpu_set_t cpuset; // typedef struct { ULong bits[CPU_SETSIZE / NCPUBITS]; }
   // new CPU pinning method for more than 32 cores
   struct s_pin {int ncpu;int ofs;int dist;int size;}; // new2013-05-07
   // most important value first (numcpus, then offset, then distance,...)
   int  ncores=0, ii; unsigned long cmask=0L;
   struct s_pin pin = {16,0,1,1}; // CFG_CPUSET=16,0,1,1 mask=0xffff
   // struct s_pin pin = {8,1,4,2}; // mask=0x00066
   for (i1=1;i1<argn;i1++) { /* check all args */
      if (strncmp(argv[i1],"--pin_ncpu=",11)==0) {
         pin.ncpu=atoi(argv[i1]+11); continue; }
      if (strncmp(argv[i1],"--pin_ofs=",10)==0) {
         pin.ofs= atoi(argv[i1]+10); continue; }
      if (strncmp(argv[i1],"--pin_dist=",11)==0) {
         pin.dist=atoi(argv[i1]+11); continue; }
      if (strncmp(argv[i1],"--pin_size=",11)==0) {
         pin.size=atoi(argv[i1]+11); continue; }
   }

   if (pin.ncpu<1) {printf("ERROR: pin.ncpu<1\n");exit(1);}
   if (pin.size<1) {printf("ERROR: pin.size<1\n");exit(1);}
   if (pin.dist<0) {printf("ERROR: pin.dist<1\n");exit(1);}
   if (pin.ofs>=pin.ncpu) {printf("ERROR: pin.shift\n");exit(1);}

   // auto-detect num cores by try+fail sched_affinity
   CPU_ZERO(&cpuset); 
   for(j1=0,i=CPU_SETSIZE;i;i>>=1){
     CPU_SET(j1|i,&cpuset);
     if (!sched_setaffinity(0, sizeof(cpu_set_t), &cpuset)) j1|=i; // no_err!
     CPU_CLR(j1|i,&cpuset);
   }
   if (!mpi_i) printf("# Max-detected-CPU-Cores/Threads= %d\n", j1+1);
   ncores=j1+1;
      
   for (ii=0;ii<8*sizeof(long) && ii<mpi_n;ii++) {  // set cmask
     j1=((ii/pin.size)*pin.dist
        +(ii%pin.size)+pin.ofs)%pin.ncpu;
     if (j1<8*sizeof(long)) cmask |= 1UL<<j1;
   }
   if (!mpi_i)
     printf("# CPU-Pinning: idx=((i/%d)*%d+(i%%%d)+%d)%%%d, mask=0x...%08lx\n",
     pin.size, pin.dist, pin.size, pin.ofs, pin.ncpu,
      cmask);
   if (pin.ncpu>ncores && !mpi_i) printf("WARNING: pin.ncpu>ncores\n");

#if 0 // old 32bit/64bit limited to cpus=bits-1
   { int in=0;
     // 2013-05 endless loop for 32 Cores on int32-mask CPUSET=0xffffffff
     // for (ii=0; CFG_CPUSET>>ii && ii<64; ii++) if (1&(CFG_CPUSET>>ii)) ncores++;
     for (ii=0; CFG_CPUSET>>ii && ii<64; ii++) if (1&(CFG_CPUSET>>ii)) {
       if (mpi_i%ncores==in) break; in++; // mycore=0,1,2,..,0,1,2,.. --byslot
     }
   }  // ii = cpucore(mpi_i)
#else  // new method for more than 32 cores
   // does only cover equidistant distribution but most
   // MASK     
   // 0xffff = {16,1,1,0}
   // 0x6666 = {16,2,4,1}
   // 0x0f0f = {16,4,8,0} = ((ii/4)*8+0)%16
   // 0xf0f0 = {16,4,8,4}
   ii=((mpi_i/pin.size)*pin.dist
      +(mpi_i%pin.size)+pin.ofs)%pin.ncpu;
   // ncores=pin.ncpu;
#endif
   if (ii>=CPU_SETSIZE) {printf("ERROR: ii>CPU_SETSIZE, ii=%d\n",ii);exit(1);}
   CPU_ZERO(&cpuset);   // void CPU_ZERO(cpu_set_t *set); bits[idx]|=mask
   CPU_SET(ii,&cpuset); //  ->bits[ii/NCPUBITS] |= (1<<(ii%NCPUBITS))
   
   // The constant CPU_SETSIZE (1024) specifies a value one greater than  the
   //      maximum CPU number that can be stored in a CPU set.
   usleep((mpi_i%16+1)*10000L); // 10ms..170ms, sort outputs
   if (mpi_i<4 || mpi_i>=mpi_n-2) {
     printf("# set cpu affinity for core(%d)=%d size=%d\n",
              mpi_i, ii, CPU_SETSIZE);
   }
   // no sched_setaffinity on OSF1
   if (sched_setaffinity(0, sizeof(cpu_set_t), &cpuset)) {
     printf("# set cpu affinity for core(%d)=%d size=%d failed\n",
             mpi_i, ii, CPU_SETSIZE);
             usleep(100000L);perror("set_affinity");
   }
   usleep(100000); // 100ms, pseudo sync
  }
  MPI_Barrier(MPI_COMM_WORLD);
#endif

  if (!mpi_i) printf("#");
   /* 2013-05 gcc -fopenmp -v ... =
     -L/usr/lib/gcc/x86_64-redhat-linux/4.1.2 -lgomp -lpthread */
#ifdef _OPENMP
  // omp_set_num_threads(2);
#pragma omp parallel
  omp_n= omp_get_num_threads();  /* must be called within parallel pragma */
  /* next line is outside parallel region */
  if (!mpi_i) printf(" OpenMP threads=%d", omp_n); /* same on every node? */
#endif
#ifdef MPI_VERSION
  if (!mpi_i) printf(" MPI%d.%d",MPI_VERSION,MPI_SUBVERSION);
#endif

  if (!mpi_i)
    printf(" mpitasks=%d mem/rank=%.1fMB int%d mintime/s=%.1f test%d\n"
           "# measure: BW = bandwidth, aBW = aggregate BW, lat = latency\n",
      mpi_n, (1l<<(log2size-10))/1024., 1<<log2long, tmin, testmode);

  if (!mpi_i) {
    printf("# CPU"
#ifdef __GNUC__
     "-gcc-%d.%d.%d"
#endif
     "-features:"
#ifdef __MMX__   // 8 registers mmx=64bit  1f64,2f32
     " MMX"  // pentium-mmx pentium2 c3 (register:MM0..MM7 64bit=1i64,2i32,4i16,8i8)
      // PADDB, PADDD, PADDSB, PADDSW, PADDUSB, PADDUSW, PADDW, 
      // PMADDWD, PMULHW, PMULLW
#endif
#ifdef __SSE__  // 8registers(IA64=16) xmm=128bit, 2 operands (float only, 2*f64,4*f32)
     " SSE"  // pentium3  register xmm0..xmm7 128bit=4f32 
     // ADDPS, ADDSS, MULPS, MULSS
     // float a[4],b[4]  a[]+=b[] 4ADDs/clk
#endif
#ifdef __SSE2__ // +integer (ToDo: check asm output by -S -msse2 -mno-sse2)
     " SSE2"  // pentium-m pentium4  XMM+2f64,16i8,8i16,4i32,2i64
     // ADDPD,MULPD 2f64 ADDSD,MUSD 1f64  (see .S output)  PADDQ PMULUDQ
     //  gcc -ftree-vectorizer-verbose=2 -ftree-vectorize -O2
     //  gcc -ftree-vectorizer-verbose=2 -O3 
#endif
#ifdef __SSE3__
     " SSE3"  // complex-arith ADDSUBPD ...PS 
#endif
#ifdef __SSSE3__
     " SSSE3"  // atom core2  PMULHRSW, PMADDUBSW
#endif
#ifdef __SSE4a__
     " SSE4a"  // core2++ PMULLD PMULDQ
#endif
#ifdef __3dNOW__
     " 3dNOW"   // k8  mmx+2f32  PFADD, PFMUL
#endif
#ifdef __FMA__
     " FMA"      // FMA fused multiply add xmm0..3
#endif
#ifdef __FMA4__
     " FMA4"     // FMA4 fused multiply 4 args r0=r1+r2*r3
#endif
#ifdef __AVX__  // 16 registers ymm0..ymm15 je 256bit 4f64,8f32
     " AVX"     // 256bit, 3 operands, needs OS-support a[]=b[]+c[] GCC4.6
    // only low128bit for integer usable!
    //  __m256, __m256d und __m256i   _mm256_*
#endif
#ifdef __AVX2__  // ??? +integer
     " AVX2"     // 256bit-SSE-iOPs, 256bit for integers, +FMA3(3register)
#endif
#ifdef __AVX512F__  //  register zmm0..31 256bit 8f64,16F32 +ymm16..31 
     " AVX512F"  // noFMA?
#endif
    "\n"
#ifdef __GNUC__
     , __GNUC__ ,  __GNUC_MINOR__, __GNUC_PATCHLEVEL__
#endif
    );
  }

  if ((memhole0 || memhole1) && (memhole1<=memhole0 || memhole0<0)){
    printf("ERROR: memhole end<=start or start<0\n");exit(1);}

  if (memhole0 || memhole1){
    long int i1; 
    static char *a1=NULL, *a2=NULL;
    printf("# reserve memory hole (start-end=%ld-%ld)\n", memhole0, memhole1);
    if (memhole1<=memhole0 || memhole0<0){
      printf("ERROR: memhole end<=start or start<0\n");exit(1);}
    a1 = (char *) malloc(          memhole0 <<20);
    a2 = (char *) malloc((memhole1-memhole0)<<20);
    if ((!a1) || !a2) {printf("ERR malloc\n");exit(7);}
    // touch every kB
    for (i1=0;i1<          memhole0 <<10;i1++) a1[i1<<10]=0; // good mem
    for (i1=0;i1<(memhole1-memhole0)<<10;i1++) a2[i1<<10]=0; //  bad mem
    if (a2[0]!=a1[0]) printf("ERROR bad mem is bad, but reserved\n");
    free(a1); // get good mem back, but do not release a2
    printf("# hole reserved, ready for progress ...\n");
    // hopefully compiler does no garbage collection!
  }
                         
  /* old: asize = 1l << log2size */
  /* new: asize = option.asize << log2size 2013-06-14 */
  aa = (long int *) malloc(asize); /* malloc but touch it by threads! */
  { /* MPI_Allreduce !aa and get min/max maxmalloc, also sync 2013-06 */
    int aa1,aa2; long int size1, size2;
    aa1=((aa)?0:1); /* !0 = failure */
    MPI_Allreduce(&aa1, &aa2, 1, MPI_INT, MPI_SUM, MPI_COMM_WORLD);
    if (aa2 && !mpi_i) printf("ERROR: malloc failed on %d nodes\n", aa2);
    if (aa1 && aa2<8) printf("INFO: malloc failed on node %d\n", mpi_i);
    if (aa2) { free(aa); aa=NULL; /* get max alloc ToDo! */
      for (size1=0,size2=asize;size2>=1<<20;size2>>=1) { /* log2-test */
        aa = (long int *) malloc(size1+size2);
        if (aa) { size1+=size2; free(aa); aa=NULL; }
      } /* now we have the maxalloc on each node, range = 1MB..2*asize-1 */
      MPI_Allreduce(&size1, &size2, 1, MPI_LONG, MPI_MIN, MPI_COMM_WORLD);
      if (!mpi_i) printf("INFO: min maxmalloc/MB = %ld\n", size2>>20);
      if (aa2<8 && size2==size1) printf("INFO: min on node %d\n", mpi_i);
    }
  }
  /* in MPI failed malloc cause signal 11 ? on CentOS */
  /* in MPI Abort generates lot of error lines Gentoo1.12.11.1 + Slurm1 */
  if (!aa) testmode=0; /* MPI_Abort(MPI_COMM_WORLD, 133); */
  
#if 0
  /* ToDo: compute CPU clock, diff loops over n vs. n+1 ops? */
  if (!mpi_i) {
    unsigned long i0,i1=0,i2,j=7,i6=1000000; double t2,t1,t0,
      tmin2=1.0; /* measuring time in seconds (old: tmin2=tmin) */
    if (argn==7000) i6=100;  /* knocking out the optimizer :) */ 
    /* ToDo: second ident loop for deviation mesurements (t1b,t2b) */ 
    t0=MPI_Wtime();t1=0.0;
    for (i0=1;i0;i0<<=1) { /* double loops until 1s gone, avoid overrun i0 */
      for (;i1<i0;i1++)
        for (i2=0;i2<i6;i2++) { j=j^i2; }
      t1=MPI_Wtime()-t0;
//      printf("# loop1=   %6lde6 / %6.3f s = %9.0fe6/s\n", i0, t1, i0/t1);
      if (t1>tmin2) break;
    }
    printf("# loop1=   %6lde6 / %6.3f s = %9.0fe6/s ^\n", i0, t1, i0/t1);
    t0=MPI_Wtime();t2=0.0;
    for (i1=0;i1<i0;i1++)   // add single ops until time is +10% minimum ??
      for (i2=0;i2<i6;i2++) { j=j^i2^3; } // MIPS-500MHz 256e6/(1.6s-1.1s)
    t2=MPI_Wtime()-t0;
    printf("# loop2=   %6lde6 / %6.3f s = %9.0fe6/s ^^\n", i0, t2, i0/t2);
    if (t2-t1<0.1) { /*  2013-06-18 not for MIPS! */
     t0=MPI_Wtime();t2=0.0;
     for (i1=0;i1<i0;i1++)   // add single ops until time is +10% minimum ??
       for (i2=0;i2<i6;i2++) { j=j^i2^(i2>>7)^3; } // AMD ^,>> parallel!
     t2=MPI_Wtime()-t0;
     printf("# loop2=   %6lde6 / %6.3f s = %9.0fe6/s ^^(>>)\n", i0, t2, i0/t2);
    }
    if (t2!=t1)
    printf("# loop2-1= %6lde6 / %6.3f s = %9.0fe6/s\n", i0,t2-t1,i0/(t2-t1));
    if (j==863479817) fputs("# super hit :)\n",stdout);
    /* 2013-05
     * Intel E5400      2.70GHz loop2-1=  32768e6 /  3.415 s =  9596e6/s ??
     * AMD Opteron 6134 2.30GHz loop2-1=  65536e6 / 28.534 s =  2297e6/s 
     * SC MIPS ICE9B    500MHz  loop2-1=    256e6 /  1.613 s =   159e6/s ??
     * ToDo: L1-cache-clock? via i=rnd[i] im cache? bzw. Rekursion(AES?)
     */
  }
#endif

#if 1
  if (testmode&(16^8^4^2)) {
    /* preinit a_rnd and reuse for speedup, but randomness gets more bad */
    srandom(mpi_i);  // different access patterns on different CPUs
    /* long int random() returns 31bit random only! */
    r00= 0l; /* xor sum */
    for (k=0;k<1024;k++) /* 64bit by 3*31 bit 2013-06-20 */
      r00 ^= a_rnd[k] = random() ^ ((random() ^ (random()<<8))<<31);
    a_rnd[1023] ^= r00;  /* ensure sum_xor a_rnd[k] = 0 */
    r00= 0l; for (k=0;k<1024;k++) r00 ^= a_rnd[k]; /* xor sum */
    if (r00!=0) printf("ERR: rnd sum^!=0 %016lx\n", r00);
    /* replace random() by (a_rnd[i&1023]^a_rnd[(i>>10)&1023]^...>>20...) */
    /* 1024^2 is not enough randomness, for 2^30 size arrays ToDo! */
  }
#endif

  if (testmode&(4^2)) {
    if (!mpi_i)
      puts("# Initialize+check random array for each MemSize (slow) ...");
  }

#ifndef NA      /* use option -DNA=5 to modify inner loops */
#define NA (8)  /* vector array */
#endif

#if 1
/* ToDo: prefetchable random access, randomness from extra 1k array + xor */
if (testmode&2){ /* prefetchable random access */
  /* register? */ long int ra[NA+4]; // replaces old r00,r01,r10,r11 2014-06
  /* 
   *  meggie: 8*Quad-AMD-Opteron-8378-2411MHz-512kB 8*32GB sse2 3dnow
   *    sse4a 3dnowprefetch(prefetch 32B to L1)
   *     mpiexec -np 32 ./memspeed --log2size=32 --pin_ncpu=32 --test=2 --pin_dist=1
   *   1pref 6.40GB/s=8B/1.25ns_64k  25.5MB/s=8B/314ns_4GB 31.3MB/s_1GB
   *   4pref 24.6GB/s    0.33ns_64K  52.9MB/s=8B/151ns_4GB 82.6MB/s_1GB
   *   8pref 34.2GB/s=8B/0.23ns_32k  57.2MB/s=8B/140ns_4GB 111MB/s_1GB
   *  10pref 34.2GB/s/c  0.23ns_32K  47.5MB/s/c  168ns_4GB 52.4MB/s_1GB mpi=32
   *    max_rndmem:  0.5*256GB/1.25GB/s= 4GB/47.5MB/s = 84s
   * AMD Opteron 6134 2.30GHz 16*4GB aBW.r[i]4p=2200MB/s,59ns r[i]=1000MB/s,127ns
   *  pse:   2*6core-iX5650-2.66GHz = 2*3 Speicherkanaele(DDR3-1333) max 2*32GB/s
   *    18*16GB-DDR3-1066MHz(a8.5GB/s) = 3 DIMMs/Channel max 51GB/s (ToDo)
   *    aBW=29.5GB/sRW  
   *    old=  11.6GB/s=8B/0.69ns(256) .. 113MB/s=8B/70.6ns(1GB)
   *    1pref  619MB/s    1.31ns(256) .. 113MB/s    70.6ns(1GB)
   *    2pref 12.2GB/s    0.65ns      .. 203MB/s    39.4ns
   *    3pref 18.4GB/s    0.44ns      .. 291MB/s    27.5ns
   *    4pref 24.5GB/s    0.33ns      .. 374MB/s    21ns
   *    6pref 24.5GB/s    0.33ns      .. 528MB/s    15ns
   *    8pref 20.5GB/s    0.39ns      .. 672MB/s    12ns
   *    16pre 19.5GB/s    0.41ns      .. 872MB/s    9.2ns       mpi=1 +best+
   *    32pre 15.0GB/s    0.53ns(L1)  .. 838MB/s    9.5ns=18*L1clk
   *    1pref  5.9GB/s/c  1.37ns      .. 108MB/s/c  74ns(256MB*4) mpi=4
   *    4pref 23.4GB/s/c  0.34ns      .. 333MB/s/c  24ns(256MB*4) mpi=4
   *    6pref 23.4GB/s/c  0.34ns      .. 445MB/s/c  18ns          mpi=4
   *    8pref 23.4GB/s/c  0.34ns      .. 442MB/s/c  18ns          mpi=4
   *    16pre 18.7GB/s/c  0.43ns      .. 440MB/s/c  18ns          mpi=4
   *    6pref 23.4GB/s/c  0.34ns      .. 400MB/s/c  20ns          mpi=6
   *    6pref 11.7GB/s/c  0.69ns      .. 218MB/s/c  37ns          mpi=6 c0-5 ??
   *    8pref 11.7GB/s/c  0.69ns      .. 252MB/s/c  32ns          mpi=6
   *    8pref 23.4GB/s/c  0.34ns      .. 366MB/s/c  22ns          mpi=6 c0-5
   *    6pref 11.6GB/s/c  0.69ns      .. 185MB/s/c  43ns -40%!    mpi=12 ~pinned
   *    8pref 23.4GB/s/c  0.34ns      .. 307MB/s/c  26ns +best+   mpi=12 pinned
   *                                     307MB/s/c(64MB-1GB(-4GB))
   *                                     286MB/s/c  28ns    _12GB mpi=12 pinned
   *    8pref 23.4GB/s/c  0.34ns      .. 528MB/s/c  15ns          mpi=6  pinned 2i
   *    8pref 23.4GB/s/c  0.34ns      .. 308MB/s/c  26ns          mpi=6  pinned 0..5
   *    max_rndmem:  3.68GB/s(12*256MB) per 288GB-Node !!! aber 288GB/12/2=12GB/core
   *       12GB/core for 0.5Mem/RndSpeed=144GB/3.43GB/s=42s
   *    numactl --physcpubind=0-5 --localalloc # 0,2,4,6,8,10 = L1speed/4? HT!
   *       trotzdem nicht stabil! Faktor ca. 2, inside-cpu-pinning geht!
   *  mkautz sicortex MIPS ICE9B 500MHz mpicc -O3 -funroll-loops == -O2
   *    old=   960MB/s=8B/8.33ns(16k) .. 38.1MB/s=8B/210ns(256MB)
   *    1pref  960MB/s    8.34ns(16k) .. 37.9MB/s    211ns(256MB)
   *    1pref  960MB/s    8.34ns(16k) .. 35.7MB/s    224ns(256MB*6) mpi=6
   *    2pref  1.22e+09   6.58ns      .. 43.1MB/s    185ns(256MB)
   *    2pref  1.22e+09   6.58ns      .. 40.5MB/s    198ns(256MB*6) mpi=6
   *    4pref  1.17e+09   6.84ns      .. 43.8MB/s    183ns(256MB)
   *    4pref  1.17e+09/c 6.84ns      .. 43.6MB/s    183ns(256MB*6) mpi=6
   *    8pref  1.09e+09   7.32ns      .. 46.5MB/s    172ns(256MB)
   *    8pref  1.09e+09/c 7.32ns      .. 42.7MB/s/c  187ns(256MB*6) mpi=6
   *    8pref  773MB/s    10.34ns     .. 34.8MB/s    230ns  --gnu -O3
   *                    L1-latency  5200*43MB/s=223GB/s
   *      0.5*Mem/speed=2GB/6/42.7MB/s = 8s
   */
  for(k=0;k<NA;k++) ra[k]=0l;
  if (!mpi_i)
    printf("# mpi omp Mem/Task[B] loops t[s]  aBW[B/s] BW[B/s] lat[ns] m=2\n");
  /* if start with maximum, you have to wait do long for first result
   *   so its better to start with minimum and reuse smaller init (ToDo)  */
   // for (a2=alen;a2>>5;a2>>=1) // asize<<(log2size-log2long+3)
   for (a2=alen;!(a2&3);a2>>=1);  // startvalue around 4*long
   for (;a2<=alen;a2<<=1) { /* 2014-06-26 start with 8*128kB + go up */
    // its twicky to have a random "next_index" table without shortcuts 
    // we need a 2nd array for prev_index
    t1=MPI_Wtime();
#pragma omp parallel for default(shared) private(k)
    for (k=0;k<a2;k++) {       // define a ring, backwards as starting point
      aa[k]=(k+a2-1)%(a2); // successors (old: &(a2-1) only for 2^n) *long*
    }
#pragma omp parallel for default(shared) private(k,i,r00,r01,r10,r11,r02,r12)
    for (k=0;k<a2;k++) {  
      // shaking the chain, this is slow! a2=arraysize (reuse older runs)
      // chooze a random pair r0=i, r1=j to exchange i+1 and j+1
      i= a_rnd[k&1023]^a_rnd[(k>>10)&1023]^a_rnd[(k>>20)&1023]; /* random() */
      r00=k;       r10=((unsigned long)i)%a2; // predecessors i, j
      r01=aa[r00]; r11=aa[r10];            // exchange partners i+1, j+1
      r02=aa[r01]; r12=aa[r11];            // successors i+2, j+2
      if (r12==r01 || r10==r01) continue;  // dont exchange neighbours
      // idex:    0   1   2   3   4   5   6
      // succs:   1   2   3   4   5   6   0  0-1-2-3-4-5-6-0
      //        r00 r01 r02 r10 r11 r12
      // exchng:      ^           ^          0-4-2-3-1-5-6-0 (1,4 exchanged)
      // succs:   4   5   3   1   2   6   0
      //          ^   ^       ^   ^   
      aa[r00]=r11; aa[r01]=r12; // new successors of i, i+1
      aa[r10]=r01; aa[r11]=r02; // new successors of j, j+1
      if (alen<=64 && !mpi_i) check_ring(a2,1,mpi_i);
      if (alen<=64 && !mpi_i) printf(" exchange: %2d %2d", (int)r01, (int)r11);
    }
    if (alen<=128*1024 && !mpi_i) check_ring(a2,0,mpi_i);
    t2=MPI_Wtime()-t1;
    MPI_Bcast(&t2, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
    if (a2>=alen/4 && t2>0 && !mpi_i) // final ini message if bigger 2s
    printf("#%3d %2d %8.2e %9ld %5.2f %8.2e %8.2e %6.2f rnd_ini.r[i]\n",
       mpi_n, 0+1*omp_n, a2*1.*sizeof(long int),
          1l, t2, mpi_n*a2*sizeof(long int)*4./t2,
                        a2*sizeof(long int)*4./t2, t2*1e9/(4.*a2));
//    printf("OK %d\n",__LINE__);

    // measure random access = sizeof(long)/memory_latency
    MPI_Barrier(MPI_COMM_WORLD);
    loops=0;
    r00=0; r01=a2/4; r10=a2/2; r11=a2/4*3;
    for (k=0;k<NA;k++) ra[k]=a2/NA*k; /* start indizes distributed */
    t1=MPI_Wtime();
    for (i=1;i<(1l<<30);i*=2) {
      j=i;
      // ToDo: bad usage if reset ll for every inner loop!?
      do { // repeat until some seconds are gone
         ll=0; k=a2;             // gcc-4.1 x86 2 instr.
// ToDo: OMP, ra[NN], NN=1 ???
         do {
           // 4 independent random access sequences
           // ToDo: use ra[NA] instead of r00,r01,...
#if 0 // old pse1GB=19.6ns
           r00=aa[r00];  // independent values, 3 of 4 are prefetchable?
           r01=aa[r00];  // independent values, 3 of 4 are prefetchable?
           r10=aa[r10];  // independent values, 3 of 4 are prefetchable?
           r11=aa[r11];  // independent values, 3 of 4 are prefetchable?
#else // new vectorized pse1GB=21.5ns
#if   NA==1
           ra[0]=aa[ra[0]];  // no value prefetchable
#elif NA==2                  // independent values, 1 of 2 are prefetchable?
           ra[0]=aa[ra[0]];  ra[1]=aa[ra[1]];  
#elif NA==3                  // independent values, 2 of 3 are prefetchable?
           ra[0]=aa[ra[0]];  ra[1]=aa[ra[1]];  ra[2]=aa[ra[2]];  
#elif NA==4                  // independent values, 3 of 4 are prefetchable?
           ra[0]=aa[ra[0]];  ra[1]=aa[ra[1]];  
           ra[2]=aa[ra[2]];  ra[3]=aa[ra[3]];  
#elif NA==5                  // independent values, 4 of 5 are prefetchable?
           ra[0]=aa[ra[0]];  ra[1]=aa[ra[1]];  
           ra[2]=aa[ra[2]];  ra[3]=aa[ra[3]];  ra[4]=aa[ra[4]]; 
#elif NA==6
           ra[0]=aa[ra[0]];  ra[1]=aa[ra[1]];  
           ra[2]=aa[ra[2]];  ra[3]=aa[ra[3]];  
           ra[4]=aa[ra[4]];  ra[5]=aa[ra[5]];  
#elif NA==7
           ra[0]=aa[ra[0]];  ra[1]=aa[ra[1]];  
           ra[2]=aa[ra[2]];  ra[3]=aa[ra[3]];  
           ra[4]=aa[ra[4]];  ra[5]=aa[ra[5]];  ra[6]=aa[ra[6]];
#elif NA==8
           ra[0]=aa[ra[0]];  ra[1]=aa[ra[1]];  
           ra[2]=aa[ra[2]];  ra[3]=aa[ra[3]];  
           ra[4]=aa[ra[4]];  ra[5]=aa[ra[5]];  
           ra[6]=aa[ra[6]];  ra[7]=aa[ra[7]];  
#elif NA==16
           ra[0]=aa[ra[0]];  ra[1]=aa[ra[1]];  
           ra[2]=aa[ra[2]];  ra[3]=aa[ra[3]];  
           ra[4]=aa[ra[4]];  ra[5]=aa[ra[5]];  
           ra[6]=aa[ra[6]];  ra[7]=aa[ra[7]];  
           ra[8]=aa[ra[8]];  ra[9]=aa[ra[9]];  
           ra[10]=aa[ra[10]];ra[11]=aa[ra[11]];
           ra[12]=aa[ra[12]];ra[13]=aa[ra[13]];
           ra[14]=aa[ra[14]];ra[15]=aa[ra[15]];
#elif NA==32
           ra[0]=aa[ra[0]];  ra[1]=aa[ra[1]];  
           ra[2]=aa[ra[2]];  ra[3]=aa[ra[3]];  
           ra[4]=aa[ra[4]];  ra[5]=aa[ra[5]];  
           ra[6]=aa[ra[6]];  ra[7]=aa[ra[7]];  
           ra[8]=aa[ra[8]];  ra[9]=aa[ra[9]];  
           ra[10]=aa[ra[10]];ra[11]=aa[ra[11]];
           ra[12]=aa[ra[12]];ra[13]=aa[ra[13]];
           ra[14]=aa[ra[14]];ra[15]=aa[ra[15]];
           ra[16]=aa[ra[16]];ra[17]=aa[ra[17]];
           ra[18]=aa[ra[18]];ra[19]=aa[ra[19]];
           ra[20]=aa[ra[20]];ra[21]=aa[ra[21]];
           ra[22]=aa[ra[22]];ra[23]=aa[ra[23]];
           ra[24]=aa[ra[24]];ra[25]=aa[ra[25]];
           ra[26]=aa[ra[26]];ra[27]=aa[ra[27]];
           ra[28]=aa[ra[28]];ra[29]=aa[ra[29]];
           ra[30]=aa[ra[30]];ra[31]=aa[ra[31]];
#else
           // does unroll work? ToDo: gcc 4.2.1 vs. 4.8.1
           int k3;
           for (k3=0;k3<NA;k3++) ra[k3]=aa[ra[k3]];  // -funroll-loops
#endif
#endif
           // ll=aa[ll];             // gcc-4.1 x86 1 instr. + latency
           // __builtin_prefetch(aa+l);
         } while(--k);            // gcc-4.1 x86 2 instr.
      } while(--j);               // gcc-4.1 x86 2 instr.
      loops+=i;  // long int = 64bit
      t2=MPI_Wtime()-t1;
      MPI_Bcast(&t2, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
      if ((t2>=tmin || t2>10 || loops>>30) && !mpi_i) // alive and final message
      printf("%c%3d %2d %8.2e %9ld %5.2f %8.2e %8.2e %6.2f %drnd i=r[i]\n",
         ((t2>=tmin)?' ':'#'), mpi_n, 1+0*omp_n, a2*1.*sizeof(long int),
         loops, t2, mpi_n*a2*sizeof(long)*1.*NA*loops/t2,
                          a2*sizeof(long)*1.*NA*loops/t2, 
                               t2*1e9/(a2*1.*NA*loops), NA);
      // l must be used, some compilers remove the mem access else
      if (t2>=tmin) break;
    }
    for (k=1;k<NA;k++) ra[0]^=ra[k];
    if (r00==r01 && r10==r11) puts("??\n"); /* dummy usage */
    if (ra[0]==17) puts("??\n"); /* dummy usage */
  } /* for (a2=alen;a2>>5;a2>>=1) down to 2**5=32B */
}


if (testmode&4){ /* non-prefetchable random access */
  if (!mpi_i)
    printf("# mpi omp Mem/Task[B] loops t[s]  aBW[B/s] BW[B/s] lat[ns] m=4\n");
  // for (a2=alen;a2>>5;a2>>=1) 
  for (a2=alen;!(a2&3);a2>>=1);  // startvalue around 4*long
  for (;a2<=alen;a2<<=1) { /* 2014-06-26 start with 8*128kB + go up */
    // its twicky to have a random "next_index" table without shortcuts 
    // we need a 2nd array for prev_index
    for (k=0;k<a2;k++) {       // define a ring, backwards as starting point
      aa[k]=(k+a2-1)%a2; // successors
    }
    for (k=0;k<a2;k++) {  // shaking the chain, this is slow!
      // chooze a random pair r0=i, r1=j to exchange i+1 and j+1
      i= a_rnd[k&1023]^a_rnd[(k>>10)&1023]^a_rnd[(k>>20)&1023]; /* random() */
      /* long % 2M may result to 0xffffffff8768 for negative longs */
      r00=k;       r10=((unsigned long)i)%a2;  // predecessors i, j
      r01=aa[r00]; r11=aa[r10];            // exchange partners i+1, j+1
      r02=aa[r01]; r12=aa[r11];            // successors i+2, j+2
      if (r12==r01 || r10==r01) continue;  // dont exchange neighbours
      // idex:    0   1   2   3   4   5   6
      // succs:   1   2   3   4   5   6   0  0-1-2-3-4-5-6-0
      //        r00 r01 r02 r10 r11 r12
      // exchng:      ^           ^          0-4-2-3-1-5-6-0 (1,4 exchanged)
      // succs:   4   5   3   1   2   6   0
      //          ^   ^       ^   ^   
      aa[r00]=r11; aa[r01]=r12; // new successors of i, i+1
      aa[r10]=r01; aa[r11]=r02; // new successors of j, j+1
      if (alen<=64 && !mpi_i) check_ring(a2,1,mpi_i);
      if (alen<=64 && !mpi_i) printf(" exchange: %2d %2d", (int)r01, (int)r11);
    }
    if (alen<=128*1024 && !mpi_i) check_ring(a2,0,mpi_i);

    // measure random access = sizeof(long)/memory_latency
    MPI_Barrier(MPI_COMM_WORLD);
    loops=0;
    t1=MPI_Wtime();
    for (i=1;i<(1l<<30);i*=2) {
      j=i;
      do { // repeat until some seconds are gone
         ll=0; k=a2;             // gcc-4.1 x86 2 instr.
         do {
           ll=aa[ll];             // gcc-4.1 x86 1 instr. + latency
           // __builtin_prefetch(aa+l);
         } while(--k);            // gcc-4.1 x86 2 instr.
      } while(--j);               // gcc-4.1 x86 2 instr.
      loops+=i;
      t2=MPI_Wtime()-t1;
      MPI_Bcast(&t2, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
      if ((t2>=tmin || t2>10) && !mpi_i) // alive and final message
      printf("%c%3d %2d %8.2e %9ld %5.2f %8.2e %8.2e %6.2f random i=r[i]\n",
         ((t2>=tmin)?' ':'#'), mpi_n, 1+0*omp_n, a2*1.*sizeof(long int),
         loops, t2, mpi_n*a2*sizeof(long)*1.*loops/t2,
                          a2*sizeof(long)*1.*loops/t2, t2*1e9/(loops*1.*a2));
      // l must be used, some compilers remove the mem access else
      if (t2>=tmin) break;
    }
  } /* for (a2=alen;a2>>5;a2>>=1) down to 2**5=32B */
}  /* non-prefetchable random access */

  r00=ll; // avoid optimizations which skip the loop

  /* 
   * AMD Opteron 6134 2.30GHz 16*4GB lstream.aBW=46.7GB/s,2.7ns p4=50GB/s,2.54ns
   * pse
   *    6*11.6GB/s=6*8B/0.69ns(256k) ..  6*2.96GB/s=6*8B/2.70ns(256MB) 1pre
   *    6*23.0GB/s=6*8N/0.34ns(16k)  ..  6*4.19GB/s=6*8B/1.93ns(256MB) 4pre
   *    6*23.0GB/s=6*8N/0.34ns(16k)  ..  6*2.47GB/s=6*8B/3.24ns(256MB) 8pre --pin_ncpu=6
   *    6*23.3GB/s=6*8N/0.34ns(16k)  ..  6*4.89GB/s=6*8B/1.64ns(256MB) 8pre --pin_ncpu=12 --pin_dist=2
   *   12*23.3GB/s=6*8N/0.34ns(16k)  .. 12*2.46GB/s=6*8B/3.26ns(256MB) 8pre --pin_ncpu=12
   * SC-MIPS-ICE9B-500MHz = 4B/clk/c (??? ToDo: how to get 2DFLOP/clk=16B/clk/c?)
   *    6*1.91GB/s=6*8B/4.18ns(16k) .. 6*349MB/s=6*8B/22.9ns(256MB) -O3 4pre
   *    6*1.91GB/s      4.20ns      .. 6*361MB/s      22.1ns        -O3 8pre
   *    6*1.73GB/s      4.61ns      .. 6*340MB/s      23.6ns        -O3 1pre
   * J-PC E5400-2.70GHz clock=(6,8,10,13.5)*200MHz=(1.2,1.6,2.0,2.7GHz)
   *    L1d=32KB-8way/core L2=2MB-8way-shared
   *   latency L1=3 L2=16clks 64B-prefetching 8-99W TDP=65W
   *   BIOS: 400MHz DDR2-800 CAS=6-6-6-18 single-channel peak=6.4GB/s lat=15ns PC2-6400
   *    lt. wiki DDR2-800 CL5-5-5-15 = 12.5ns 37.5ns, approx 6-6-6-18 15ns..45ns
   *   memtest86 1*2GB(1c?) L1=44GB/s L2=19GB/s mem2GB=3GB/s (zu wenig? 32bit? 2.7G)
   *   memspeed.i486-1.2GHz.V2 rnd4=4B/128ns prnd2=4B/73ns 4B/3.20ns=1.25GB/s
   *   memspeed.m32-1.2GHz.V32 rnd4=4B/128ns prnd2=4B/73ns 4B/1.94ns=2.06GB/s
   *   memspeed.m32-2.7GHz.V16 rnd4=4B/105ns prnd2=4B/37ns 4B/1.02ns=3.91GB/s
   *   memspeed.m64-1.2GHz.V4  rnd4=8B/126ns prnd2=8B/47ns 8B/1.95ns=4.10GB/s
   *   memspeed.m64-1.6GHz.V4  rnd4=8B/115ns prnd2=8B/45ns 8B/1.72ns=4.66GB/s
   *   memspeed.m64-2.0GHz.V4  rnd4=8B/111ns prnd2=8B/43ns 8B/1.63ns=4.92GB/s
   *   memspeed.m64-2.7GHz.V4  rnd4=8B/105ns prnd2=8B/42ns 8B/1.48ns=5.40GB/s
   *    gcc492 -m64 -DNA=1 -ftree-vectorize -march=native -funroll-loops -Ofast
   *     1.6GHz: 256MB=5.0GB/s L2=11.5GB/s L1=22.5GB/s (ToDo: disasm!)
   *     random-blocksize=256MB (latenz steigt propto blocksize) L1prnd=clk
   *   1/1200MHz=0.833ns/clk gcc492,NA=2,O2,native: 16kB=2*8B/0.84ns
   *   ToDo: CPU auf 2*FSB=1600MHz setzen!?
   * 
   */
if (testmode&1){ /* sequential read = lstream */
  MPI_Barrier(MPI_COMM_WORLD);
  if (!mpi_i)
    printf("# mpi omp Mem/Task[B] loops t[s]  aBW[B/s] BW[B/s] lat[ns] m=1\n");
  // measure memory throughput for step0
  for (a2=alen;a2>>7;a2>>=1) { // reduce msgsize to 1<<7=128 * 8 = 1k..2kB
    long int *ap=aa;  // moving pointer    
    long int x1, x2, x3, x4; /* decoupled 4 subresults */
    long int ra[NA+4]; // replaces old r00,r01,r10,r11 2014-06
    //    double t3;
    loops=0; x1=x2=x3=x4=0;
    for(k=0;k<NA;k++) ra[k]=0l;
#pragma omp parallel for default(shared) private(k)
    for (k=0;k<a2;k++) {       // initialize mem (using thread affinity)
      aa[k]=-(k&1);            // 00 ff pattern
    }
    t1=MPI_Wtime();
    for (i=1;i<(1l<<30);i*=2) { // repeat until mintime
      j=i;
      do { // repeat until some seconds are needed
         long int *ae=aa+a2;  // 2013-10-30
         // echo 1200000 >/sys/dev*/system/cpu/cpu0/cpufreq/scaling_max_freq
         //  k=a2/4;do{ ... }while(--k); # t=1.92ns(256M) 0.84(10k) 0.88(1k)
         //  for(k=0;k<a2;k+=4){ ... };  # t=1.95ns(256M) 0.84(10k) 0.88(1k)
         // omp1(k=0;k<a2;k+=4){ ... };  # t=1.96ns(256M) 1.20(16k) 6.64(1k)
         //  manual unroll has no effect on 2700MHz, but 1200MHz???
         //  for(k=0;k<a2;k+=4) # t=1.54ns(256M) PC2700MHz
         //  for(k=0;k<a2;k+=1) # t=1.51ns(256M) 0.58(32k) 0.70(1k) PC2700MHz
         //  for(k=0;k<a2;k+=1) # t=1.97ns(256M) 1.24(32k) 1.57(1k) PC1200MHz
         // ToDo: ... -Os
#pragma omp parallel for default(shared) private(k,ap) reduction(^:x1,x2,x3,x4) firstprivate(aa,a2)
// ToDo: all threads start with ap = bad! missing const directive a2
// lot of cpu overhead (16-32x) for openmp + small arrays (small a2)! cache coh.
//   -Os besser als -O2 !?
// NA==1 slower on cache (fast ram)
         for (k=0;k<a2;k+=NA) {  /* down=2.08-2.13ns, up=2.16-2.20ns */
#if   NA == 1
           ra[0]^=aa[k]; // 1.51ns(256M,PC2700MHz) 1.97ns(256M,PC1200MHz)
                      // 1.97ns(8M) 1.53ns(1M) 1.22ns(16k) gcc44.Os=O3 PC1200
                      //   erreicht bei L1-cache 0.58ns = 1.566clks = ...
#elif NA == 2  
           ra[0]^=aa[k  ]; ra[1]^=aa[k+1]; 
#elif NA == 3
           ra[0]^=aa[k  ]; ra[1]^=aa[k+1]; ra[2]^=aa[k+2];
#elif NA == 4  // mem-bustakt? DDR800(400MHz)*4B? memtest=2996MB/s 4B/1.54ns=2.6GB/s
           ra[0]^=aa[k  ]; //     0.833ns = 1/1200MHz (5/4*0.833ns=1.04ns)
           ra[1]^=aa[k+1]; // 1.96ns(8M) 1.53ns(1M) 0.84ns(16k) gcc44.O3=Os PC1200
           ra[2]^=aa[k+2]; // 1.96ns(gcc44,256M,PC1200MHz -Os=O2=O3=gcc41.O3)
           ra[3]^=aa[k+3]; // 1.54ns(256M,PC2700MHz) 2.07ns(256M,PC1200MHz -Os)
           // gcc44O3openmp 1task: 1.96ns(256M) 4.1GB/s 1.20ns(16k)  6ns(1k) 1200MHz
           // gcc44O3openmp 2task: 1.79ns(256M) 4.5GB/s 1.18ns(16k) 12ns(1k) 1200MHz
           //  Beschleunigung kaum sichtbar! warum nicht? mem=3GB/s L1=44GB/s
           //    mem long/BW=8B/3000MB/s=2.6ns  nax=800MHz*8B=6.4GB/s
           // openmp overhead ab 16k sichtbar, unter 4k dominant
#elif NA == 8
           ra[0]^=aa[k  ]; ra[1]^=aa[k+1]; ra[2]^=aa[k+2]; ra[3]^=aa[k+3];
           ra[4]^=aa[k+4]; ra[5]^=aa[k+5]; ra[6]^=aa[k+6]; ra[7]^=aa[k+7];
#else 
           int k3;
           for (k3=0;k3<NA;k3++) ra[k3]^=aa[k+k3];  // -funroll-loops
#endif
//   nicht besser mit pointer ap=aa+k, da openmp int braucht, hier nicht genutzt
// __builtin_prefetch(aa); // gcc-4.1 x86-msse2 2 instr.
// schlechter fuer kleine arrays wegen coherence tests!
         }
         /* ToDo: j loop is bad for x1^x2... result! */
      } while(--j);               // gcc-4.1 x86 2 instr.
      loops+=i;
      t2=MPI_Wtime()-t1;
      xy=0; for(k=0;k<NA;k++) xy ^= ra[k]; // x1^x2^x3^x4;
      if (xy!=0) { /* vor MPI_Reduce! */
        printf("ERR: bad result x=%016lx %s\n", xy,
         getenv("SLURMD_NODENAME")); // no HOSTNAME
        // system("hostname"); // 2013-06-18 kautz, failes wenn OOM
      }
      /* Bcast laeuft auch durch L2-ECC-gecrashte kautz-nodes! 2013-06 */
      MPI_Bcast(&t2, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD); /* old */
      /* MPI_Reduce bringt komische Resultate fuer loops, t2 passt nicht? */
      // t3=t2;MPI_Reduce(&t3, &t2, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD);
      if ((t2>=tmin || t2>10) && !mpi_i) // alive and final message
      printf("%c%3d %2d %8.2e %9ld %5.2f %8.2e %8.2e %6.2f lstreamV%d*%dB\n",
         ((t2>=tmin)?' ':'#'), mpi_n, omp_n, a2*1.*sizeof(long int),
         loops, t2, mpi_n*a2*1.*sizeof(aa[0])*loops/t2,
                          a2*1.*sizeof(aa[0])*loops/t2,
                t2*1e9/(loops*1.*a2),NA,(int)sizeof(aa[0])); /* aa=long_int=8B*/
      // l must be used, some compilers remove the mem access else
      if (t2>=tmin) break;
    } /* repeat until mintime */
  }
} /* sequential read */

/* ToDo: STRESS = endless write/read (count writes, reads) report bitflips */
/* numademo+memtest86+v4.20 is bad stresstest because single thread
 *   using mpirun -np 16 numademo? 
 *  8: 00    00    00    00   vs. ff    ff    ff    ff 
 * 16: rnd rnd^~0 rnd rnd^~0  vs. rnd^~0 rnd rnd^~0 rnd
 */
if (testmode&(8^16)){
  /* new alloc to allow locality for OMPI (could be done at begin!) */
  /* ToDo OMP, old: ompi only, use double + FPU for PSU test
   * comp2+kW? + check result (store first to source and compare later)
   * comp2: 2*(8*2300MHz,64GB)  numademo4n8GB: rnd=90MB/s memcpy=2-6GB/s 200m++
   *   numademo wandert zw. cores, use taskset 0x01 ... ? wird ignoriert!
   *     numactl --physcpubind 0 wird ab running on node 1 ignoriert!
   *   numademo  8m stream random #  24s str=3-6GB/s rnd=200-300MB/s
   *   numademo 16m stream random #  57s str=3-5GB/s rnd=170-260MB/s
   *   numademo 80m stream random #      str=3-5GB/s rnd=150-220MB/s
   *  mpicc -O2 -o memspeed -DCFG_MPI -DCFG_CPUSET=0xffff ../utils/memspeed.c 
   *  mpirun -np 16 ./memspeed --log2size=30 --mintime=2 --test=8 # 
   * 2013.05: 16*1GB 16*2925MB/s (no cpuset), 16*3760MB/s (cpuset16) 60GB/s
   *          16*2GB 16*3751GB/s (37GBcached == drop_caches)
   *          16*4GB 16*3740GB/s # node04: 3737MB/s, node19: 
   */
  int tmode=16, cnt=0;
  if (!mpi_i)
    printf("# R/W+Power-Stress + check for bitflips (mode=%d)\n"
   "# mpi omp Mem/Task[B] loops t[s]  aBW[B/s] BW[B/s] lat[ns]\n", testmode);
#pragma omp parallel for default(shared) private(k) firstprivate(aa,alen)
  for (k=0;k<alen;k++) aa[k]=-(k&1);   // initialize mem 00 ff pattern
  // measure memory throughput for step0
  for (cnt=0,a2=alen;;cnt++) { // do not reduce msgsize, stress it endless! alen=1<<alog
    long int x1, x2, *ap=aa, ae=0l, ae2;  // moving pointer
    /* rot through bit 8 + 16 until 1 bit found */
    for(;;) {
      tmode=((tmode<<1)&0x18) | ((tmode&0x10)>>1);
      if (tmode&testmode) break;
    }

    ae=a_rnd[cnt&1023]^a_rnd[(cnt>>10)&1023];  /* rnd(1024^2), start=0 */
    if (tmode&8) ae=0l; /* 00 ff sequence */
    
    loops=0; x1=x2=xy=0;
    t1=MPI_Wtime();
    for (i=1;i<(1l<<30);i*=2) {
    register double r1=0.12, r2=0.12, r3=0.13, r4=0.14; // FLOPS to stress PDU
    register double r5=0.22, r6=0.22, r7=0.23, r8=0.24;
    double r9=0.31;
      j=i; // 2013 comp2: first round runs slower, autotune cache method?
      // 1st_round=393MB/s, 2nd++_round=1610MB/s (faktor4), training? throttle?
      do { // repeat until some seconds are reached
         ap=aa; k=a2/2; ae^=~0l;
         ae2= ((tmode&16)?ae^~0l:ae); // gcc-4.1 x86 2 instr.
         // k=a2/2;do{...}while(--k);  # 8.49ns(256MB) 0.94GFLOPs(19.6%)
         // for(k=0;k<a2;k+=2){...}    # 8.69ns(256MB) 0.92GFLOPs(19.2%)
         // omp1(k=0;k<a2/2;k++){...}  # 4.52ns(256MB) 1.77GFLOPs(36.9%) 1.2GHz
         // 1200MHz*4FLOP=4.8GFLOPs (8FLOP/1.67ns) 
        // Pentium(R) Dual-Core CPU E5400 @ 1200-2700MHz
// ToDo: bad pragma-option for r1-r8 (reduction is bad)
// ToDo: ap[] access is all threads to same memory == bad!
#pragma omp parallel for default(shared) private(k) reduction(^:x1,x2) firstprivate(aa,a2,ae,ae2) reduction(+:r1,r2,r3,r4,r5,r6,r7,r8)
         for(k=0;k<a2;k+=2) {  // 2013-09-11 // 3 instr.
//         do {
//           ap=aa+k;
// omp.L330 8*mulsd %xmmX,%xmmX + 8*addsd %xmmY, %xmmX (16 instr.) 2FLOPs/clk
// gcc.L207 8*(2*movsd+mulsd+addsd) from/to memory     (32 instr.) 1FLOP/clk
// ToDo: vergleich auf moderner CPU!
           r1=r1*r1+0.1;    // cpu burn code (stress the system) 8 FLOPs/cycle
           r2=r2*r2+0.2;  // ToDo: use array r[3] as r3 and later unroll-loops
           r3=r3*r3+0.3;
           r4=r4*r4+0.4;
           r5=r5*r5+0.1;
           r6=r6*r6+0.2;
           r7=r7*r7+0.3;
           r8=r8*r8+0.4; // without next 4 lines 8FLOP/0.21ns=38GFLOPs @ 1.2GHz?
//                 gcc-O1 8FLOP/5.44ns=1.47GFLOPs(31%*4.8GF) (faster!?)
//                 gcc-O2 8FLOP/7.10ns=1.13GFLOPs(23%*peak4.8GF)
//                 omp1-O2 8FLOP/6.91ns=1.16GFLOPs (shared rx)
//                 omp1-O2 8FLOP/3.34ns=2.40GFLOPs(50%*4.8GF) (reduction rx)
//          t2=MPI_Wtime()+MPI_Wtime(); // localisation in .s
#if 0
           x1^=*(ap  );           // gcc-4.1 x86 1 instr. + latency
           x2^=*(ap+1);
           *(ap  )=ae;
           *(ap+1)=ae2;           //   ... new pattern write
           ap+=2;                 // gcc-4.1 x86 1 instr., bad for openmp!
           // gcc44-O3 4.57ns PC1200MHz
#else 
           x1^=aa[k];           // gcc-4.1 x86 1 instr. + latency
           x2^=aa[k+1];
           aa[k  ]=ae;
           aa[k+1]=ae2;           //   ... new pattern write
           // gcc44-O3 4.52ns PC1200MHz (4.58ns 2 threads! slower?!)
           //  read + write max reached?  3.5GB/s(r+w)
           // quanta2: 9GB/s(1) 18.2GB/s(2) 36GB/s(4) (68GB/s)(2.6GB/8)
           //     104GB/s(4GB/16)  97GB/s(6GB/32) 96GB(80GB/64)
           // OMP_NUM_THREADS=16 nice -19 ./memspeed --test=8 --asize=750000
           //   750GB 7.11. 18:18 ECC-Fehler! ^c
           //   370GB 18:19 ECC-Fehler
           //   200GB 18:20 ECC-Fehler
           //   100GB 18:21 ECC-Fehler 104GB/s, 50GB 18:21:54
           //  10GB 82.9GB/s ohne Fehler 16MB cache
           //  20GB 104GB/s mit Fehlern 18:25:14
           //  15GB  81GB/s mit Fehlern 18:25:54
           //  12GB 105GB/s mit Fehlern 18:26:55
           //    11GB 105GB/s mit Fehlern 18:27:42
           //    10GB 105GB/s mit Fehlern 18:28:25
           
           //     8GB 103GB/s mit Fehlern 18:29:25
           //     # ToDo einzelne CPUs testen (numactl) ! works test=8
           //   numactl --hardware # --physcpubind 0-7 --membind=0 26.7GB/s
           //     1cpu=9GB/s(remote=5.0GB/s,remote2=3.8GB/s) 4nodes=105GB/s
           //  ToDo: --memhole=10G-11G (alloc in 1GB + touch, free)
           //      --test=idle
#endif
           // __builtin_prefetch(aa); // gcc-4.1 x86-msse2 2 instr.
           // ToDo: better prefetch every 4th access only + unroll loop
//         } while(--k);            // gcc-4.1 x86 2 instr.
         } // for(k=0..a2/2)
         // ToDo: problem x1 x2 ^= over multiple loops hides bitflips!
      } while(--j);               // gcc-4.1 x86 2 instr.
      loops+=i;
      t2=MPI_Wtime()-t1;
      MPI_Bcast(&t2, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
      if ((t2>=tmin || t2>60*60) && !mpi_i) // alive and final message
      /* aBW= 1.02e+06 %8.2e 1%, 2013-06-20 fix bad factor 2 Mem/Task */
      printf("%c%3d %2d %8.2e %9ld %6.2f %8.2e %8.2e %6.2f lr+w4*+%02d\n",
         ((t2>=tmin)?' ':'#'), mpi_n, omp_n, a2*1.*sizeof(long int),
         /* factor 2 for r+w */
         loops, t2, mpi_n*a2*2.*sizeof(long int)*loops/t2, /* r+w = 2x */
                          a2*2.*sizeof(long int)*loops/t2,
                t2*1e9/(loops*1.*a2), tmode);
      // comp2: 2*1.6GB/s = 2*8B*200MHz=2*8B*2.3GHz/11.5clks (7used?)
      // l must be used, some compilers remove the mem access else
      xy=x1^x2;
      if (xy!=0lu)
//       printf("WARNING: xy= %016lx bitflip?\n", xy);
        printf("ERR: bad result rnd=%016lx x=%016lx %s\n", ae, xy,
         getenv("SLURMD_NODENAME")); // no HOSTNAME
      if (r1==0.6 || r2==0.4 || r3==0.5 || r4==0.23) printf("??\n");
      if (r9==3.4 || r8==0.6 || r5==0.4 || r6==0.5 || r7==0.2) printf("??\n");
      if (t2>=tmin) break;
    }
  }
}


#endif

// TODO also include gather/load + scatter/store vector-benchmarks test64?
// TODO        a[i]+=b[ind[i]]*c[i] or a[ind[i]]+=b[i]*c[i]
// TODO    as SpMV core micro-benchmark

// testmode&32  == max-FLOPs-only use registers only
//       NV=vecSize(AVX=8*f64), NU=alu-units(1-2 per core)?
//       NR=usable_registers (maxAVX2=16,avx512=32)
//       NI=independetOPs = (NR-1)^2 ??? compute + output
// SuperMUC-NG master 2*Xeon(R) Gold 6148 2.40GHz fam6 model85 20cores
//             CPU_incl.Omnipath 2 AVX512-FMA-units = 32 FLOP/cycle
// SuperMUC-NG nodes  2*Xeon(R) Platinum 8174 3.10GHz fam6 model85 48cores
//             but show 2.3GHz (8176=28c) 2xAVX512FMA mem6ch 8wayL1
// SuperMUC-NG mpicc uses gcc480 nur SSE2, mpiicc (icc license only on master)
//  gcc48 -O2 -g -march=native -mtune=native -ftree-vectorize -funroll-loops
//    -S -o - | grep -e vfmadd.\*[yz]mm # vfmadd231sd xmm, + some ymm-vOPs
//  gcc82 -O2 -g -march=native -mtune=native -ftree-vectorize -funroll-loops
//    -S -o - | grep -e vfmadd.\*[yz]mm # vfmadd231sd ymm, no zmm-Ops AVX512F
//     28.6GF/2401MHz=12F/cycle  vs. 32DP/c AVX512=8DP FMA=2OPs/clk 2units?
//  gcc73 -O2 NV8  NP4       = vfmadd132pd zmm 55.4GF/clk=20.5F*2.4GHz
//  gcc73 -O2 NV16 NP4       = vfmadd132pd zmm 81.8GF/clk=30.3F*2.4GHz
//  gcc73+icc19 -O2 NV16 NP4 = vfmadd231pd ymm 57.0GF/clk=23.8F*2.4GHz
//  gcc73+icc18 -O2 NV16 NP4 = vfmadd231pd ymm 57.1GF/clk=23.8F*2.4GHz 3*8F-ports?
// jPC 2cP-E5400-2.7GHz(1.2GHz)-2MB ssse3 vmx 
//  gcc4.1.2 -O3 -funroll-all-loops -mssse3  1.26 GF32  1.0/cycle
//  gcc4.4.7 -O3 -funroll-all-loops -mssse3  2.39 GF32  2.0/cycle
//  gcc4.4.7 -O3 -funroll-loops -DNV=20      7.28 GF32  6.1/cycle
//  gcc4.8.2 -O3 -funroll-loops -DNV=20      7.99 GF32  6.7/cycle
// PC16 i7-6700-4c3.4GHz-2fma-avx2 gcc8    4*27.1 GF64 7.97/cycle*3.4GHz NA4 NP4
// PC16 i7-6700-4c3.4GHz-2fma-avx2 gcc8    4*32.6 GF64 9.59/cycle*3.4GHz NA4 NP5
// PC16 i7-6700-4c3.4GHz-2fma-avx2 gcc8    4*39.0 GF64 11.5/cycle*3.4GHz NA8 NP3
//   lt. foren 2(!) AVX-FMA-units(je 8FLOPs/clk) = 16 Flops/cycle
//   hpl.txt  WR00C2R2.N12800.NB128.2x2=150.7GFLOPs 4*oblaso = 69% 
// comp2 2s*8c-AMD_Opteron_6134 2300MHz 512kB 3dnowext sse4a ht
//  gcc4.1.2 -O3 -funroll-loops -DNA=8       3.48 GF-SP 1.51/cycle a+=b*c
//  gcc4.4.0 -O3 -funroll-loops -DNA=8       4.57 GF-SP 1.99/cycle
//  gcc4.4.0 -O3 -funroll-loops -DNA=25      11.8 GF-SP 5.13/cycle maxSP>4
//  gcc4.4.0 -O3 -funroll-loops -DNA=4..16   4.59 GF-DP 2.00/cycle maxDP=2
//  gcc4.4.0 -O3 -funroll-loops -DNA=20      7.19 GIOPs 3.13/cycle ???
// pse2012-1n*2s*6c-X5650-2.67GHz(1.60-3.06GHz) Westmere sse2 popcnt
//   12*2.67GHz*4F=12*10.7GF/c=127GF LinP=114GF-DP
//   sse(4f32=4F32/cyc) sse2(2f64,2i64,4i32,8i16,16i8,8F32/cyc*+,not fused!)
//   ssse3(permutes bytes in word) ~avx(256bit,3operands=16F32/cyc*+fused)
//  gcc4.4.6 -O2 -DNA=8                      5.25 GF    2.0/cycle
//   -O3 -DNA=20 -funroll-loops              17.2 GF-SP 6.4/cycle max?
//   -O3 -DNA=8..16                          6.12 GF-SP 2.3/cycle =2*3.06GHz!
//   -O3 -DNA=8..16                          6.12 GF-DP 2.3/cycle a+=b*c[]
//   -O3 -DNA=8..10 -funroll-loops           6.12 GF-SP 2.3/cycle a*=b; a+=c;
//   -O3 -DNA=8..10 -funroll-loops           6.12 GF-DP 2.3/cycle a*=b; a+=c;
// see linux.doc GCC-4.8.1
// quanta2014 E5-4620 + http://linux.web.cern.ch/linux/devtoolset/ gcc-4.8.2
//     gcc4.8.2 -O3 -DNA=32 -funroll-loops -mavx # 8 FLOP/clk*2.2GHz E5-4620
//            NA=32 15.2GF  7F/clk-DP
//            NA=64 34.9GF 16F/clk-SP            # 16 FLOP/clk == peak
// meggie2009 Opt8378-512kB-2411MHz*4F gcc4.8.1 -O2    # 0.5 F64/clk*2.4GHz
// meggie2009 Opt8378-512kB-2411MHz*4F gcc4.8.1 -Ofast # 2   F64/clk*2.4GHz
//     2008 Shanghai K10 L1=256KB-code+256kB-data L2=2M L3=6M 3dNow! noFMA!
//           SSE2, SSE3, SSE4a + VM  2.7F/clk
//           SSE2: (4DP-FLOPs/cycle,8SP-FLOPs/cycle)
//     (-O2 -DNA=8 -funroll-loops) # 4FLOPs not reached! no float-advant.
//     (16B=2DP=4SP)/cycle/corepair=4FMAC/corepair= 4FLOP/clk/core
//   # gcc4.8.1 -Ofast == -O3 -ffast-math SP=19.3GF/clk (testrange NA=2..40)
//   gcc4.1.2 -O3 -ffast-math 10 2.22 GF-SP
//   gcc4.8.1 -march=amdfam10 -O3 NA=4..16 4.81 GF-SP = 2F32 (wrong arch?)
//   gcc4.8.1 -march=amdfam10 -O3 NA=28    9.60 GF-SP = 4F32
//   gcc4.8.1 -march=amdfam10 -O3 NA=56    9.62 GF-SP = 4F32
//   gcc4.8.1 -march=k8-sse3  -O3 NA=56    9.62 GF-SP = 4F32
//   gcc4.8.1 -march=k8       -O3 NA=56    9.62 GF-SP = 4F32
//   gcc4.1.2 -march=k8       -O3 NA=56    1.46 GF-SP  # == amdfam10
//   gcc4.1.2 -march=amdfam10 -O3 NA=8     2.42 GF-SP = 1F32 
//   gcc4.1.2 -march=amdfam10 -O3 NA=8     2.45 GF-SP = 1.02F32 -ffast-math
//   gcc4.6.3 -march=amdfam10 -O3 NA=8     4.81 GF-SP = 2F32    -ffast-math
//   gcc4.6.3                 -Ofast NA=10 4.81 GF-SP
//   gcc4.6.3 -march=amdfam10 -O3 NA=20    5.63 GF-SP = 2.34F32
//   gcc4.7.2 -march=amdfam10 -Ofast NA=10 4.81 GF-SP ???
//   gcc4.7.2                 -Ofast NA=10 19.2 GF-SP !!!
//   gcc4.8.1 -Ofast NA=8,12,16 4.82 GF = 2    SP-FLOPs/clk (SP,DP)
//   gcc4.8.1 -Ofast NA=18      5.84 GF = 2.42 SP-FLOPs/clk (SP,DP)
//   gcc4.8.1 -Ofast NA=20      6.41 GF = 2.66 SP-FLOPs/clk (SP,DP)
//   gcc4.8.1 -Ofast NA=6       14.5 GF = 6    SP-FLOPs/clk  75% a=a+b*c
//   gcc4.8.1 -Ofast NA=5 ?     19.3 GF = 8    SP-FLOPs/clk 100% a=a+b*c sse2 !!!
//   gcc4.8.1 -Ofast NA=10,14   19.3 GF = 8    SP-FLOPs/clk 100% a=a+b*c sse2 !!!
//   gcc4.8.1 -O3    NA=4-5     4.82 GF = 2    SP-FLOPs/clk  25% a=a+b*c
//   gcc4.8.1 -Ofast NA=5 ?     3.00 GF = 1.24 SP-FLOPs/clk  16% a=a*b+c,a=(a+b)*c
//   no double 4 FLOP found!
//   /opt/x86_open64-4.5.2/bin/opencc-4.5.2 -Ofast -DNA=8 -funroll-all-loops
//      1.84 GF = 0.83 FLOP/clk ???  a=(a*b)+c (2.0 GF fuer a=a*b+c)
// Atom-N450-1.66GHz-512kB(1.0GHz) 8.5W + 2.5W/7.6GF sse2 ssse3
//   gcc4.7.2 detects MMX SSE SSE2
//   gcc4.7.2 -Ofast -DNA=5 -funroll-loops   # 7.6  SP-GF/clk 1GHz a=a+b*c
//   gcc4.7.2 -Ofast -DNA=6                  # 4.66 SP-GF/clk 1.0GHz
//   gcc4.7.2 -Ofast -DNA=8             # 1.63-1.84 SP-GF/clk 1.0GHz SSE2
//   gcc4.7.2 -Ofast -DNA=8 -mssse3     # 1.72-1.91 SP-GF/clk 1.0GHz SSSE3
//   gcc4.7.2 -Ofast -DNA=10                 # 6.50 SP-GF/clk 1.0GHz SSE2
//   gcc4.7.2 -Ofast -DNA=14                 # 6.87 SP-GF/clk 1.0GHz SSE2
//   gcc4.7.2 -Ofast -DNA=14 -funroll-loops  # 7.72 SP-GF/clk 1.0GHz a+=b*c
//   gcc4.7.2 -Ofast -DNA=14 -funroll-loops  # 7.77 SP-GF/clk 1.0GHz a+=b*c -mno-sse2
//       16* fmul %st(0), %st; faddl .LC78(%rip);fxch %st(5) # i+=32
//   gcc4.6.3 -Ofast -DNA=14 -funroll-loops  # 1.73 SP-GF/clk 1.0GHz a+=b*c
//    andere 3 viel langsamer! wieso?
//   gcc4.7.2 6*xmm-Register (128bit=4*f32) 2*3mem128write+xmm0=12
// MAD = multiply add a*b+c
// FMA = fused multiply add a*b+c with no round step for product
// MAC.FMA a=a+(b*c)  not on opteron?
//  FMA3: $0=$0*$2+$1 $0=$0+$1*$2(skalar-prod)
//  FMA4: $0=$1*$2+$3
// sse: 4 indpendend f32 ops at same time (SIMD) for(i=0;i<3;i++)a[i]+=b[i];
//  http://www.linuxjournal.com/content/introduction-gcc-compiler-intrinsics-vector-processing
//  /usr/local/lib/gcc/x86*/4.8.1/include/mmintrin.h - X86 MMX (64bit vector)
// 3DNow (AMD deprec.) = packed-2xfloat32-SIMD/VECT  2..4FLOPs +AVG8i8
// 3dnowprefetch = prefetch 32B to L1, multithread.monitor+mwait
// sse2=SIMD-int 128bit(4*SP) parallel add + mul(=2*) = 8SP or 4DP
// sse3=complex, add(mltply) 2f64 or 2*2f32 in same reg. (=3DNow) 2..4FLOP
// sse4a=LZCNT, POPCNT,   
// Intel Core 2 and Nehalem, SSE:      4 DP FLOPs/cycle ( 8 SP FLOPs/cycle)
// Intel Sandy Bridge/Ivy Bridge, AVX: 8 DP FLOPs/cycle (16 SP FLOPs/cycle)
// Intel Haswell with AVX2/FMA:       16 DP FLOPs/cycle (32 SP FLOPs/cycle)
// AMD K10 (Opteron, Athlon etc) SSE2: 4 DP FLOPs/cycle ( 8 SP FLOPs/cycle)
// AMD K11/Bulldozer with FMA:         8 DP FLOPs/cycle (16 SP FLOPs/cycle)
// IBM Power 7:                        8 DP FLOPs/cycle (16 SP FLOPs/cycle)
//
//  matrix*matrix to reuse elements! C=A*B or C=A*A + scalarprod
//             b00 b01
//             b10 b11
//    a00 a01  a00*b00+a01*b10  2*2=8*mul+4add(dep)  A+=C(4add)
//    a10 a11    
//    A*A+B und C*C+B interleaved 2MA2x2=16mul+16add, MA3x3=27mul+27add
//                                    8*F64=64B          18*8=144B
//    max reuse of vector?
//      a0*a0 ... a3*a3  b0+c0 ... b3+c3  3v4-regs 2v4instr v=vector
//      a0+c0 ... a3+c3  b0*b0 ... b3*b3  3v4-regs 2v4instr  16FLOP/2clk NA=8
//    + repeat (2*vec4
// ToDo: INT-OPs?
//  SSE independent 4*= + 4+=
//    ToDo: autotest best (ex: kautz.MIPS-ICE9B 2FLOPs 74% by r11*=r11;r10+=ry)

/* FLOP_t may be defined outside as float,double,int,short,char ... */
#ifdef FLOP_t
#  if FLOP_t != float && FLOP_t != double && FLOP_t != short
#    error "FLOP_t must be one of float, double or short" 
#  endif
#endif
#ifndef FLOP_t  /* double precission (DP) OR single precision (SP, 32bit) */
#define FLOP_t float  // better define float-size 4 or 8
#endif
#ifndef NR // number of SIMD-registers usable
# if   defined(__AVX512F__)
#   define NR 32 /* zmm0..zmm31 */
# elif defined(__AVX__) || defined(__SSE__) // AVX,SSE,MMX? = 16 registers
#   define NR 16 /* ymm0..ymm15, xmm0..xmm15 */
# else   // MMX and lower
#   define NR 8  /* xmm0..xmm7 */
# endif
#endif  // NR SIMD-registers
#ifndef NV // width of SIMD-registers 
# if   defined(__AVX512F__)
#   define NV 16 /* 16*f32=512 */
# elif defined(__AVX__)
#   define NV  8 /*  8*f32=256 */
# elif defined(__SSE__)
#   define NV  4 /*  4*f32=128 */
# else 
#   define NV 2  /*  2*f32=64 */
# endif
#endif  // NR SIMD-registers
#ifndef  NI
# define NI 5  // 5 clocks fma-latency, independend OPs for filled pipeline
#endif // NI
/* ToDo: FLOAT64-STRESS = max power consumption 2013-09-12 */
//#error  ToDo NA=32 but 4x parallel f32/i32 SSE 4 flop/clk
if (testmode&32) {  // float precision may double peak performance
      // NV needed for vectorization 1..32
      // MP is for pipelining (independed registers), MP*NV*8 < L1=16kB,regs?
      //    MMX    =                  2*f32 or 1*f64
      //    SSE    = 16*XMM_reg, XMM= 4*f32 or 2*f64 a=a+b
      //    AVX    = 16*YMM_reg, YMM= 8*f32 or 4*f64 gcc4.6 k2.6.30 2009-06
      //    AVX512 = 32*ZMM_reg, ZMM=16*f32 or 8*f64 gcc4.9 k3.15 ? 2015?
      // PC16.RH7=k3.10  avx+avx2 (linux too old for avx512) 
      // PC16.Ubuntu18=k4.15 avx+avx2 also (no avx512, cpu is avx2 only)
      // PC16.avx2  4^2*4f64  1c: 27.1GF = 1*3.4GHz*8.0F/cycle
      // hpc18=zen.RH7=k3.10 2*(16c*2.4GHz,170W,159GB/s)
      // hpc18.avx2 5^2*4f64  1c: 9.82GF = 1*1.2GHz*8.2F/cycle master+tset2,4..
      // hpc18.avx2 5^2*4f64  1c: 19.6GF = 1*2.4GHz*8.2F/cycle master+taskset1
      // hpc18.avx2 5^2*4f64  1c: 19.8GF = 1*2.4GHz*8.2F/cycle node ??
      //  TODO master incorrect 1.2GHz clock at core 1..15, 2.4GHz at core0
      // register // does not compile, but auto compiles sometimes well
      FLOP_t ra[NI*NV];// __attribute__ ((aligned (64))); // half speed?!
      FLOP_t rb[NI*NV];// let gcc see independent arrays
      FLOP_t rc[NI*NV];// let gcc see independent arrays
      //  NR=3*NI
      // NI=independent writes?
      // FLOP_t rc[NP][NV];// __attribute__ ((aligned (64)));
      // FLOP_t rd[NP][NV];// __attribute__ ((aligned (64)));
      // 2D-koord-rotations? rot-matrix cos(x)=1-x^2/2  sin(x)=x-x^3
      //    x1[2]=rot(x1[2])+ofs[2];  4x-independend parallel, 5x indep. seq
      //  end-vector-len=? drift?
      FLOP_t rtmp[NV]; // initialize registers
      FLOP_t r0,r1,r2, ry=0.0;
  int tmode=32, cnt=0, k1,k2,k3,k4;
  // if (63&(int)(&(ra[0][0]))) printf("WARNING: ra unaligned 64B\n");
  if (NR>16 && !mpi_i)
    printf("WARNING: %d SIMD registers needed (max=16 SSE-AVX)\n", NR);
  if (NR>32 && !mpi_i)
    printf("WARNING: %d SIMD registers needed (max=32 AVX512)\n", NR);
  if (!mpi_i) printf("# SIMD size is %d*f%d (max=512 AVX512),"
   " regNR=%d indepNI=%d\n", NV, sizeof(FLOP_t)*8, NR,NI);
  if (!mpi_i) {
    { FILE *f1=NULL; unsigned long l1=0,l2=0,l3=0;
      f1=fopen("/sys/devices/system/cpu/cpu0/cpufreq/scaling_cur_freq","r");
      if (f1) { fscanf(f1,"%lu",&l1); fclose(f1); }
      f1=fopen("/sys/devices/system/cpu/cpu0/cpufreq/scaling_min_freq","r");
      if (f1) { fscanf(f1,"%lu",&l2); fclose(f1); }
      f1=fopen("/sys/devices/system/cpu/cpu0/cpufreq/scaling_max_freq","r");
      if (f1) { fscanf(f1,"%lu",&l3); fclose(f1); }
      printf("# CPUfreq.min,max,cur[MHz]: %lu %lu %lu\n",l2/1000,l3/1000,l1/1000);
    }
    printf("# AI=ArithmeticIntensity= %d FLOP / %d Byte = %.3f FLOP/Byte\n",
           (NR-NI+1)*NI*NV*2, (NR*NV*sizeof(FLOP_t)), // FLOP : Byte
           (NR-NI+1)*NI*NV*2./(NR*NV*sizeof(FLOP_t)));
    printf("# FLOAT%d-Power-Stress (mode=%d, 2DPFLOP/clk @ X5650)\n"
   "# mpi omp loops t[s]  aFLOPS[1/s] FLOPS[1/s] lat[ns]\n", 
   (int)sizeof(FLOP_t)*8, testmode);
  } // MPI
  t0=MPI_Wtime();
  // measure memory throughput for step0
#pragma omp parallel default(shared) private(k,k2) firstprivate(a2,ry,rb,rc)
{  // reduction of array (+:ra) does not work in C, need to do it by hand 2016
//    no reduction of arrays possible in C/C++ 2016-01 reduction(+:ra0,ra1...)
#pragma omp for  // reduction ry
// default(shared) private(k,k2) firstprivate(a2,ry,rb,rc)
// ... reduction(+:ra) ERROR 2016-01
  for (cnt=0,a2=alen;;cnt++) { // do not reduce msgsize, stress it endless! alen=1<<alog
    /* rot through bit 8 + 16 until 1 bit found */
    a2=64000000; /* fix for unroll, must be dividable by n*NV */
    loops=0;
#pragma omp single
    t1=MPI_Wtime();
/*
# PC16-start i7-6700 4c8t-CPU 3.40GHz +avx2 2018 minMHz=800 ca.1s to maxMHz
#   $312 14nm 2015.09 4c 8T TDP=65W MxMem=64GiB IGP=HD-Graphics530-350MHz
#   6th Generation (Skylake Microarchitecture, Skylake S, 16FLOP/cycle/core?)
#   ToDo: RPeak=217.6GFLOP/s(4c) = 54.4GFLOPs/core HPL.xhpl memtest?
#     ATTENTION! check assembler code, gcc may generate strange ASM-code
#     PC16-cpu has two FMA units (each has 4*f64 * 2FLOP/f64) = 16 FLOP/cycle
#  TODO check is AVX-FMA is supported in assembly! = yes (see fma_avx.c)
#   gcc-4.8 erzeugt 32*vfmadd231pd %ymm12,%ymm1,%ymm14 + 12*muell = 46%
#
# PC16 4c*3.4GHz*16F=217.6GF WR00C2R2.N12800.NB128.2x2=150.7GFLOPs 4*oblaso
# PC16 memspeed g82-O2+vec+NV32+f64  4*25.4GFLOP = 46%
#   4ch*8GT/s, maxmemspeed= 34GB/s (/16B * 2FLOP = 4.25GFLOP = 1.95% RPeak)
#   4 cores * 16 FLOP/clk = 4c*54.4GF/c = 217.6 GF
#   ideal Algorithmic Intensity >= 217.6 GFLOP / 34 GB/s = 6.4 FLOP/Byte
#   ideal Algorithmic Intensity >= 217.6 GFLOP / 34 GB/s = 6.4 FLOP/Byte
#      AI  >= 51.2 FLOP / input_double
#      AI  >= 51.2 FLOP / input_double
#      AI  >= 410 FLOP / 8double   # needs n^3 FLOPs over T[2,2,2]-rank3tensor
#      AI  >= 461 FLOP / 9double   # needs n^3 FLOPs over M[3,3]
#     V[N].V[N] = 2N FLOP, M[N,N].M[N,N]= 2N^3 FLOP?
#    rank1-tensor V[N]: sum over 1idx = rank0      2N-FLOP S+=V[i]
#    rank2-tensor M[N,N]: sum over 1idx = rank2    M[i,j]+= A[i,k]*B[k,j]
#    rank2-tensor M[N,N]: sum over 2idx = rank0         S+= A[k,l]*B[l,k]
#        for N=2 equivalent to complex vector[2]-skalarprod (but other signs)
#    rank3-tensor T[N,N,N]: sum over 2idx= rank2   M[i,j]+= T1[i,k,l]*T2[l,k,j]
#    rank4-tensor T[N,N,N,N]: sum over 2idx= rank4   N^4flops
#    PRNG=random.gen=  Fibonacc?  i[n]=i[n-1]*i[n-2]+0.1 
#    Mandelbrot? znew=zold^2+c (c=-1.4, -1.25, -1, -0.75, -0.5, -0.25, 0.25) 
#  mandelbrot?  cplx z=z*z+c (abs(c)<1) for every c
#   chaos-polynome, n*2-teilchen-systeme zeitentwicklung 
#  SW: rank3-tensors? (3x3x3=27 doubles)
# see https://www.7-cpu.com/cpu/Skylake.html # 2019-05
# SkyLake AVX512 registers=32*(8*f64)SIMD  ? nicht alle Skylakes? k3.10
# SkyLake L1 (32kB-data+32kB-instr)/core 64B/line 8-way latency=4..5cycles
# SkyLake L2 256kB 64B/line 4-way latency=12cycles
# SkyLake RAM latency=42cycles+51ns (2channels=0.5/core) 
# SkyLake branch misprediction penalty = 16.5-20 cycles, 64byte-cross=7cycles
    RAM Read  B/W (Parallel Random Read) = 5.9 ns / 64B-cache-line = 10800 MB/s
      # 10.89 GB/s per CPU(!) = 2 channels (latency hidden, 64B/60ns=1GB/s worst)
    RAM Read  B/W (Read, 16-64 Bytes step) = 26000 MB/s
    RAM Read  B/W (Read, 32 Bytes step - pointer chasing) = 16500 MB/s
    RAM Write B/W (Write, 8 Bytes step) = 17800 MB/s
  TODO: how much Byte can be R/W (L1 or RAM) to get max FLOPs 16FLOP/cycle/core???
       4*54.4 GFLOP/s vs. 10.8GB/s = 1280 FLOP vs. 64 Byte (20F:1B)   RAM.rnd64B
         54.4 GFLOP/s vs. ??? GB/s = ??? FLOP vs.  ??? B L1
         ## -O2 -mavx2 -mfma ? gcc485=3GB/s gcc821=11.1GB/s(1c)
  AVX512: 8*2FLOP(SIMD-FMA) vs. 8*8B(f64) = 2 FLOP : 8 Byte (1F:4B) registers
    some Skylake CPUs have 2 AVX512 units per core, most have 1 AVX512-unit 
    TODO: cplx simu?
  TODO perf stat -d ./memspeed --test=32  # bad if showing L1-dcache-loads = 5G/s 
  # perf list # L1-dcache-loads L1-dcache-stores  L1-dcache-load-misses 
  # perf stat -e fp_arith_inst_retired.scalar_double ... # FMA=2x same as output
  # perf stat -e fp_arith_inst_retired.256b_packed_double ... # *4double
  TODO https://en.wikipedia.org/wiki/Loop_nest_optimization
  TODO  nur bis NP=2 compact avx-fma-code but 32 instructions (expect 8)
  # PC16.gcc8 pragma-unroll ist bad! gcc than produces 2FMA.SSE not 1FMA.AVX
 ## AI = Arithmetic Intensity
    AI = Floor NI*(NR-NI)*NV*2 FLOP /  NR*NV*fsize register-size
  # NP = independent OPs for pipeline usage + latency hiding = NR - 2 (?)
*/
    for (i=1;i<(1l<<30);i*=2) {  // 1..1G-log2-loop
      double ry=0.0; long rnd=random();
      for (k3=0;k3<NI;k3++) { // different SIMD-registers
        for (k2=0;k2<NV;k2++) rtmp[k2]=0.9*k2/NV+0.002*((rnd>>(2*k2))&3);
        for (k2=0;k2<NV;k2++) rb[k3*NV+k2]=rtmp[k2]; // avoid register pointer error?
      }
      for (k3=0;k3<NI;k3++) { // different SIMD-registers
        for (k2=0;k2<NV;k2++) rtmp[k2]=0.9*k2/NV+0.002*((rnd>>(2*k2))&3);
        for (k2=0;k2<NV;k2++) rc[k3*NV+k2]=rtmp[k2]; // avoid register pointer error?
      }
      // TODO 2D-rotations? [x,y]=[[a00,a01],[a10,a11]]*[x,y]
      //              =      ra      rb         rc       ra
      j=i; do { // repeat until some seconds are reached
//  upto NP=2 compact avx-fma-code (OK) but 32 instructions (expect 8) TODO
//    4fold unrolled? reordered?
//  NP=3... memory-used = low perf, TODO why? 13.5GFLOP/s = 25%
//     32*SIMD: vfmadd231sd %xmm4, %xmm4, %xmm14 # SSE registers used only!
//        but ymm registers used outside loop!
//        SSE: 2f64*2FMA= 8FLOP  *3.4GHz = 27.2 GFLOP/s (PC16 2 FMA-units)
//        AVX: 4f64*2FMA=16FLOP  *3.4GHz = 54.4 GFLOP/s
//        L1=12.6*8B*3.4GHz=342GB/s(memtest)
//     2*NP register vs. NP*NP FLOP
//     RoofLine-Analysis NV*32FLOP : 12*vRegister  or L1-R/W?
//         (NR-NI+1)*NI*NV*2, (NR*NV*sizeof(FLOP_t)), // FLOP : Byte
//#pragma GCC unroll (0)  // 4*4=16avx-register
         for(k=0;k<a2;k+=(2*NI*NI*NV*2)) {  // 2013-09-11 // 3 instr.

           // FMA4 read_3double write_1double 1mult 1add = 32B 2FLOP
           // L1-speed 16kB or AVX2-register speed? 
           // GCC supports FMA4 -mfma4 since 4.5.0[21] and FMA3 -mfma since 4.7.0.
           // NR*SIMD_registers + (2+(NR-2NV))*vecFMA
           //  maxPC16 dbl NV=4 NI=6 NR=18 (16reg+2L1)  38.5GF/54.4GF
           //  maxPC16 dbl NV=4 NI=7 NR=21 (16reg+2L1)  44.5GF/54.4GF 
           //  maxPC16 dbl NV=4 NI=8 NR=24 (16reg+2L1)  46.3GF/54.4GF 
           //  PC16 NV4 NI10 NR30 double 42.3GF/54.4GF/c  78%   3.4GHz
           //  SMuc NV8 NI5  NR15 double 45.3GF/73.6GF/c        2.3GHz gcc/7 
           //  SMuc NV8 NI6  NR18 dbl     54.5GF/73.6GF/c       2.3GHz gcc/7 
           //  SMuc NV8 NI7  NR21 dbl 48c*61.0GF/73.6GF/c 83%   2.3GHz 
           //  SMuc NV8 NI8  NR24 dbl     71.8GF/73.6GF/c 98%   2.3GHz +++
           //  SMuc NV8 NI9  NR27 dbl      9.8GF/73.6GF/c 13%   2.3GHz ---asm
           //    mpicc+gcc7 -O2 -mavx512f -mfma -funroll-loops -ftree-vectorize
           //    asm: vfmadd231pd %zmm2, %zmm6, %zmm1 (45x) NI=5..8
           //   a += b*c
#pragma GCC unroll (16)  // 4*4=16avx-register, NR >= 2*NI
           for (k4=0;k4<NI;k4++) { // artificial parameters for FLOPs
             for (k3=0;k3<NI;k3++) { // fill pipeline with NI independend FLOP
               // vfmadd(ijk)pd intel=op(i) * op(j) + op(k)/mem -> op(1)
               //               AT+T=op(4-i)/mem * op(4-j) + op(4-k) -> op(3)
               for (k2=0;k2<NV;k2++)  // vector unroll
                 ra[k3*NV+k2] += rb[k3*NV+k2]*rc[((k3+k4)%NI)*NV+k2];
             } // k3 independent registers NV(tic) + NV(toc) + NR-2NV(args)
           } // k4 parameters for FLOPs
#pragma GCC unroll (16)  // 4*4=16avx-register, NR >= 2*NI
           for (k4=0;k4<NI;k4++) { // artificial parameters for FLOPs
             for (k3=0;k3<NI;k3++) { // fill pipeline with NI independend FLOP
               for (k2=0;k2<NV;k2++)  // vector unroll
                 rb[k3*NV+k2] +=  ra[k3*NV+k2]*rc[((k3+k4)%NI)*NV+k2];
             } // k3 independent registers NV(tic) + NV(toc) + NR-2NV(args)
           } // k4
         } // for(k=0..a2, k+=2) // == FLOP intense core
         // ToDo: problem x1 x2 ^= over multiple loops hides bitflips!
      } while(--j);               // gcc-4.1 x86 2 instr.
#pragma omp single
      t2=MPI_Wtime();
      if (((int)(t1))>123456789) break; // find code in asm by grep 12345
      // if (((int)(t2-t1))>123456789) break; // easier find 1234567 at .s code
      loops+=i;
      for (k2=0;k2<NV;k2++) rtmp[k2]=0;
      for (k3=0;k3<NI;k3++)
      for (k2=0;k2<NV;k2++) // reduce vectors to single vaue
        rtmp[k2] += ra[k3*NV+k2];  // critical reduction
#pragma omp critical
{
      for (k2=0;k2<NV;k2++) // reduce vectors to single vaue
        ry += rtmp[k2];   // critical reduction
}
      t3= t2 - t1;
#pragma omp single   // also synchronizes
      MPI_Bcast(&t3, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
      if ((t3>=tmin || t3>60*60) && !mpi_i) { // alive and final message
        FILE *f1=NULL; unsigned long l1=0,l2=0,l3=0;
        f1=fopen("/sys/devices/system/cpu/cpu0/cpufreq/scaling_cur_freq","r");
        if (f1) { fscanf(f1,"%lu",&l1); fclose(f1); }
        /* aBW= 1.02e+06 %8.2e 1%, 2013-06-20 fix bad factor 2 Mem/Task */
        printf("%c%3d %2d %9ld %6.2f %8.2e %8.2e %6.2f %d*+%02d %luMHz %d\n",
         ((t3>=tmin)?' ':'#'), mpi_n, omp_n,
      /* factor 4*+ ToDo: aSPEED=threads*loops/maxtime SPEED=loops/mintime */
         loops, t3, mpi_n*a2*1.*loops/(t2-t1), a2*1.*loops/((t2-t1)*omp_n),
                omp_n*(t2-t1)*1e9/(loops*1.*a2), NV, tmode, (l1+500)/1000,
                (ry<=0.0)?1:0);
      }
      if (t3>=tmin) break; // stop doubling loops
    } // i*=2  double loops each round
    if ((t2-t0)>=opt_time) break; // abort after walltime
  } // loops
} // reduction of array by hand, omp parallel section
} // testmode&32 = CPU-FLOPS

  if (!mpi_i) printf("# memspeed finished\n");
  
  MPI_Finalize();
  return xy+r00;
}
